#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML NumPy materializer."""

import os
from collections import Counter
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

import numpy as np

from zenml.enums import ArtifactType, VisualizationType
from zenml.logger import get_logger
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.metadata.metadata_types import DType, MetadataType

if TYPE_CHECKING:
    from numpy.typing import NDArray

logger = get_logger(__name__)


NUMPY_FILENAME = "data.npy"

DATA_FILENAME = "data.parquet"
SHAPE_FILENAME = "shape.json"
DATA_VAR = "data_var"

# Check NumPy version for compatibility handling
IS_NUMPY_2 = np.lib.NumpyVersion(np.__version__) >= "2.0.0"

# In NumPy 2.0, np.object_ is deprecated in favor of object
# Let's use the right type based on the NumPy version
NUMPY_OBJECT_TYPE = object if IS_NUMPY_2 else np.object_


def _ensure_dtype_compatibility(arr: "NDArray[Any]") -> "NDArray[Any]":
    """Ensure consistent dtype handling across NumPy versions.

    Args:
        arr: NumPy array to ensure compatible dtype handling

    Returns:
        NumPy array with consistent dtype behavior
    """
    if IS_NUMPY_2:
        return arr  # NumPy 2.0 already preserves precision
    else:
        # For 1.x, explicitly preserve precision when needed
        return arr.astype(arr.dtype, copy=False)


def _create_array(
    data: Any, dtype: Optional[Union["np.dtype[Any]", Type[Any]]] = None
) -> "NDArray[Any]":
    """Create arrays with consistent behavior across NumPy versions.

    Args:
        data: Data to convert to array
        dtype: Optional dtype to use

    Returns:
        NumPy array with consistent creation behavior
    """
    if IS_NUMPY_2:
        return np.asarray(data, dtype=dtype)
    else:
        # In NumPy 1.x, copy behavior is different
        return np.array(data, dtype=dtype, copy=False)


class NumpyMaterializer(BaseMaterializer):
    """Materializer to read data to and from pandas."""

    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (np.ndarray,)
    ASSOCIATED_ARTIFACT_TYPE: ClassVar[ArtifactType] = ArtifactType.DATA

    def load(self, data_type: Type[Any]) -> "Any":
        """Reads a numpy array from a `.npy` file.

        Args:
            data_type: The type of the data to read.


        Raises:
            ImportError: If pyarrow is not installed.

        Returns:
            The numpy array.
        """
        numpy_file = os.path.join(self.uri, NUMPY_FILENAME)

        if self.artifact_store.exists(numpy_file):
            with self.artifact_store.open(numpy_file, "rb") as f:
                arr = np.load(f, allow_pickle=True)
                # Ensure consistent dtype handling
                return _ensure_dtype_compatibility(arr)
        elif self.artifact_store.exists(os.path.join(self.uri, DATA_FILENAME)):
            logger.warning(
                "A legacy artifact was found. "
                "This artifact was created with an older version of "
                "ZenML. You can still use it, but it will be "
                "converted to the new format on the next materialization."
            )
            try:
                # Import old materializer dependencies
                import pyarrow as pa  # type: ignore
                import pyarrow.parquet as pq  # type: ignore

                from zenml.utils import yaml_utils

                # Read numpy array from parquet file
                shape_dict = yaml_utils.read_json(
                    os.path.join(self.uri, SHAPE_FILENAME)
                )
                shape_tuple = tuple(shape_dict.values())
                with self.artifact_store.open(
                    os.path.join(self.uri, DATA_FILENAME), "rb"
                ) as f:
                    input_stream = pa.input_stream(f)
                    data = pq.read_table(input_stream)
                vals = getattr(data.to_pandas(), DATA_VAR).values
                arr = np.reshape(vals, shape_tuple)
                # Ensure consistent dtype handling
                return _ensure_dtype_compatibility(arr)
            except ImportError:
                raise ImportError(
                    "You have an old version of a `NumpyMaterializer` ",
                    "data artifact stored in the artifact store ",
                    "as a `.parquet` file, which requires `pyarrow` for reading. ",
                    "You can install `pyarrow` by running `pip install pyarrow`.",
                )

    def save(self, arr: "NDArray[Any]") -> None:
        """Writes a np.ndarray to the artifact store as a `.npy` file.

        Args:
            arr: The numpy array to write.
        """
        # Ensure consistent dtype handling before saving
        arr = _ensure_dtype_compatibility(arr)

        with self.artifact_store.open(
            os.path.join(self.uri, NUMPY_FILENAME), "wb"
        ) as f:
            np.save(f, arr)

    def save_visualizations(
        self, arr: "NDArray[Any]"
    ) -> Dict[str, VisualizationType]:
        """Saves visualizations for a numpy array.

        If the array is 1D, a histogram is saved. If the array is 2D or 3D with
        3 or 4 channels, an image is saved.

        Args:
            arr: The numpy array to visualize.

        Returns:
            A dictionary of visualization URIs and their types.
        """
        if not np.issubdtype(arr.dtype, np.number):
            return {}

        try:
            # Save histogram for 1D arrays
            if len(arr.shape) == 1:
                histogram_path = os.path.join(self.uri, "histogram.png")
                histogram_path = histogram_path.replace("\\", "/")
                self._save_histogram(histogram_path, arr)
                return {histogram_path: VisualizationType.IMAGE}

            # Save as image for 3D arrays with 3 or 4 channels
            if len(arr.shape) == 3 and arr.shape[2] in [3, 4]:
                image_path = os.path.join(self.uri, "image.png")
                image_path = image_path.replace("\\", "/")
                self._save_image(image_path, arr)
                return {image_path: VisualizationType.IMAGE}

        except ImportError:
            logger.info(
                "Skipping visualization of numpy array because matplotlib "
                "is not installed. To install matplotlib, run "
                "`pip install matplotlib`."
            )

        return {}

    def _save_histogram(self, output_path: str, arr: "NDArray[Any]") -> None:
        """Saves a histogram of a numpy array.

        Args:
            output_path: The path to save the histogram to.
            arr: The numpy array of which to save the histogram.
        """
        import matplotlib.pyplot as plt

        plt.hist(arr)
        with self.artifact_store.open(output_path, "wb") as f:
            plt.savefig(f)
        plt.close()

    def _save_image(self, output_path: str, arr: "NDArray[Any]") -> None:
        """Saves a numpy array as an image.

        Args:
            output_path: The path to save the image to.
            arr: The numpy array to save.
        """
        from matplotlib.image import imsave

        with self.artifact_store.open(output_path, "wb") as f:
            imsave(f, arr)

    def extract_metadata(
        self, arr: "NDArray[Any]"
    ) -> Dict[str, "MetadataType"]:
        """Extract metadata from the given numpy array.

        Args:
            arr: The numpy array to extract metadata from.

        Returns:
            The extracted metadata as a dictionary.
        """
        if np.issubdtype(arr.dtype, np.number):
            return self._extract_numeric_metadata(arr)
        elif np.issubdtype(arr.dtype, np.str_) or np.issubdtype(
            arr.dtype, NUMPY_OBJECT_TYPE
        ):
            return self._extract_text_metadata(arr)
        else:
            return {}

    def _extract_numeric_metadata(
        self, arr: "NDArray[Any]"
    ) -> Dict[str, "MetadataType"]:
        """Extracts numeric metadata from a numpy array.

        Args:
            arr: The numpy array to extract metadata from.

        Returns:
            A dictionary of metadata.
        """
        # Ensure consistent precision handling
        arr = _ensure_dtype_compatibility(arr)

        min_val = np.min(arr).item()
        max_val = np.max(arr).item()

        numpy_metadata: Dict[str, "MetadataType"] = {
            "shape": tuple(arr.shape),
            "dtype": DType(arr.dtype.type),
            "mean": np.mean(arr).item(),
            "std": np.std(arr).item(),
            "min": min_val,
            "max": max_val,
        }
        return numpy_metadata

    def _extract_text_metadata(
        self, arr: "NDArray[Any]"
    ) -> Dict[str, "MetadataType"]:
        """Extracts text metadata from a numpy array.

        Args:
            arr: The numpy array to extract metadata from.

        Returns:
            A dictionary of metadata.
        """
        # Convert all array elements to strings explicitly to handle
        # mixed types and ensure NumPy 2.0 compatibility
        str_items = [str(item) for item in arr.flat]
        # Use dtype='U' (unicode string) instead of str to avoid type issues
        str_arr = _create_array(str_items, dtype=np.dtype("U")).reshape(
            arr.shape
        )

        text = " ".join(str_arr)
        words = text.split()
        word_counts = Counter(words)
        unique_words = len(word_counts)
        total_words = len(words)
        most_common_word, most_common_count = word_counts.most_common(1)[0]

        text_metadata: Dict[str, "MetadataType"] = {
            "shape": tuple(arr.shape),
            "dtype": DType(arr.dtype.type),
            "unique_words": unique_words,
            "total_words": total_words,
            "most_common_word": most_common_word,
            "most_common_count": most_common_count,
        }
        return text_metadata

    def get_item_count(self, data: Any) -> Optional[int]:
        """Get the number of items for the given data.

        Args:
            data: The data to get the number of items for.

        Returns:
            The number of items for the given data.
        """
        return len(data)

    def load_item(self, data_type: Type[Any], index: int) -> Any:
        """Load a specific item of the data.

        Args:
            data_type: The type of the data to load.
            index: The index of the item to load.

        Returns:
            The loaded item.
        """
        data = self.load(data_type)
        item = data[index]

        if not isinstance(item, data_type):
            try:
                item = data_type(item)
            except Exception as e:
                # We only log an error here, potentially pydantic can handle the
                # conversion when validating step function inputs.
                logger.error(
                    "Failed to convert item `%s` to expected type `%s`. This "
                    "is most likely due to a mismatching type annotation on "
                    "your step function input. Error: %s",
                    item,
                    data_type.__name__,
                    e,
                )

        return item



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Util functions for step and pipeline entrypoint functions."""

import inspect
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    NamedTuple,
    NoReturn,
    Optional,
    Sequence,
    Type,
    Union,
)

from pydantic import ConfigDict, ValidationError, create_model

from zenml.constants import ENFORCE_TYPE_ANNOTATIONS
from zenml.exceptions import StepInterfaceError
from zenml.logger import get_logger
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.metadata.lazy_load import LazyRunMetadataResponse
from zenml.steps.utils import (
    OutputSignature,
    parse_return_type_annotations,
    resolve_type_annotation,
)
from zenml.utils import yaml_utils

if TYPE_CHECKING:
    from zenml.config.source import Source
    from zenml.pipelines.pipeline_definition import Pipeline

    MaterializerClassOrSource = Union[str, "Source", Type["BaseMaterializer"]]

logger = get_logger(__name__)


class StepArtifact:
    """Class to represent step output artifacts."""

    def __init__(
        self,
        invocation_id: str,
        output_name: str,
        annotation: OutputSignature,
        pipeline: "Pipeline",
        chunk_index: Optional[int] = None,
        chunk_size: Optional[int] = None,
    ) -> None:
        """Initialize a step artifact.

        Args:
            invocation_id: The ID of the invocation that produces this artifact.
            output_name: The name of the output that produces this artifact.
            annotation: The output type annotation.
            pipeline: The pipeline which the invocation is part of.
            chunk_index: The index of the chunk to load.
            chunk_size: The size of the chunk to load.
        """
        self.invocation_id = invocation_id
        self.output_name = output_name
        self.annotation = annotation
        self.pipeline = pipeline
        self.chunk_index = chunk_index
        self.chunk_size = chunk_size

    def __iter__(self) -> NoReturn:
        """Raise a custom error if someone is trying to iterate this object.

        Raises:
            StepInterfaceError: If trying to iterate this object.
        """
        raise StepInterfaceError(
            "Unable to unpack step artifact. This error is probably because "
            "you're trying to unpack the return value of your step but the "
            "step only returns a single artifact. For more information on how "
            "to add type annotations to your step to indicate multiple "
            "artifacts visit https://docs.zenml.io/concepts/steps_and_pipelines#multiple-return-values."
        )


def validate_reserved_arguments(
    signature: inspect.Signature, reserved_arguments: Sequence[str]
) -> None:
    """Validates that the signature does not contain any reserved arguments.

    Args:
        signature: The signature to validate.
        reserved_arguments: The reserved arguments for the signature.

    Raises:
        RuntimeError: If the signature contains a reserved argument.
    """
    for arg in reserved_arguments:
        if arg in signature.parameters:
            raise RuntimeError(f"Reserved argument name '{arg}'.")


class EntrypointFunctionDefinition(NamedTuple):
    """Class representing a step entrypoint function.

    Attributes:
        inputs: The entrypoint function inputs.
        outputs: The entrypoint function outputs. This dictionary maps output
            names to output annotations.
    """

    inputs: Dict[str, inspect.Parameter]
    outputs: Dict[str, OutputSignature]

    def validate_input(self, key: str, value: Any) -> None:
        """Validates an input to the step entrypoint function.

        Args:
            key: The key for which the input was passed
            value: The input value.

        Raises:
            KeyError: If the function has no input for the given key.
            RuntimeError: If a parameter is passed for an input that is
                annotated as an `UnmaterializedArtifact`.
            RuntimeError: If the input value is not valid for the type
                annotation provided for the function parameter.
            StepInterfaceError: If the input is a parameter and not JSON
                serializable.
        """
        from zenml.artifacts.external_artifact import ExternalArtifact
        from zenml.artifacts.unmaterialized_artifact import (
            UnmaterializedArtifact,
        )
        from zenml.client_lazy_loader import ClientLazyLoader
        from zenml.models import ArtifactVersionResponse

        if key not in self.inputs:
            raise KeyError(
                f"Received step entrypoint input for invalid key {key}."
            )

        parameter = self.inputs[key]

        if isinstance(
            value,
            (
                StepArtifact,
                ExternalArtifact,
                ArtifactVersionResponse,
                ClientLazyLoader,
                LazyRunMetadataResponse,
            ),
        ):
            # If we were to do any type validation for artifacts here, we
            # would not be able to leverage pydantics type coercion (e.g.
            # providing an `int` artifact for a `float` input)
            return
        elif isinstance(value, list) and all(
            isinstance(item, StepArtifact) for item in value
        ):
            # Same as above, we don't do any type validation for a list of
            # artifacts here.
            return

        # Not an artifact -> This is a parameter
        if parameter.annotation is UnmaterializedArtifact:
            raise RuntimeError(
                "Passing parameter for input of type `UnmaterializedArtifact` "
                "is not allowed."
            )

        if not yaml_utils.is_json_serializable(value):
            raise StepInterfaceError(
                f"Argument type (`{type(value)}`) for argument "
                f"'{key}' is not JSON serializable and can not be passed as "
                "a parameter. This input can either be provided by the "
                "output of another step or as an external artifact: "
                "https://docs.zenml.io/user-guides/starter-guide/manage-artifacts#managing-artifacts-not-produced-by-zenml-pipelines"
            )

        try:
            self._validate_input_value(parameter=parameter, value=value)
        except ValidationError as e:
            raise RuntimeError(
                f"Input validation failed for input '{parameter.name}': "
                f"Expected type `{parameter.annotation}` but received type "
                f"`{type(value)}`."
            ) from e

    def _validate_input_value(
        self, parameter: inspect.Parameter, value: Any
    ) -> None:
        """Validates an input value to the step entrypoint function.

        Args:
            parameter: The function parameter for which the value was provided.
            value: The input value.
        """
        # We allow passing None for optional annotations that would otherwise
        # not be allowed as a parameter
        config_dict = ConfigDict(arbitrary_types_allowed=value is None)

        # Create a pydantic model with just a single required field with the
        # type annotation of the parameter to verify the input type including
        # pydantics type coercion
        validation_model_class = create_model(
            "input_validation_model",
            __config__=config_dict,
            value=(parameter.annotation, ...),
        )
        validation_model_class(value=value)


def validate_entrypoint_function(
    func: Callable[..., Any],
    reserved_arguments: Sequence[str] = (),
) -> EntrypointFunctionDefinition:
    """Validates a step entrypoint function.

    Args:
        func: The step entrypoint function to validate.
        reserved_arguments: The reserved arguments for the entrypoint function.

    Raises:
        StepInterfaceError: If the entrypoint function has variable arguments
            or keyword arguments.
        RuntimeError: If type annotations should be enforced and a type
            annotation is missing.

    Returns:
        A validated definition of the entrypoint function.
    """
    signature = inspect.signature(func, follow_wrapped=True)
    validate_reserved_arguments(
        signature=signature, reserved_arguments=reserved_arguments
    )

    inputs = {}

    signature_parameters = list(signature.parameters.items())
    for key, parameter in signature_parameters:
        if parameter.kind in {parameter.VAR_POSITIONAL, parameter.VAR_KEYWORD}:
            raise StepInterfaceError(
                f"Variable args or kwargs not allowed for function "
                f"{func.__name__}."
            )

        annotation = parameter.annotation
        if annotation is parameter.empty:
            if ENFORCE_TYPE_ANNOTATIONS:
                raise RuntimeError(
                    f"Missing type annotation for input '{key}' of step "
                    f"function '{func.__name__}'."
                )

            # If a type annotation is missing, use `Any` instead
            parameter = parameter.replace(annotation=Any)

        annotation = resolve_type_annotation(annotation)
        inputs[key] = parameter

    outputs = parse_return_type_annotations(
        func=func,
        enforce_type_annotations=ENFORCE_TYPE_ANNOTATIONS,
    )

    return EntrypointFunctionDefinition(
        inputs=inputs,
        outputs=outputs,
    )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""MLflow experiment tracker flavor."""

from typing import TYPE_CHECKING, Any, Dict, Optional, Type

from pydantic import Field, model_validator

from zenml.config.base_settings import BaseSettings
from zenml.experiment_trackers.base_experiment_tracker import (
    BaseExperimentTrackerConfig,
    BaseExperimentTrackerFlavor,
)
from zenml.integrations.mlflow import MLFLOW_MODEL_EXPERIMENT_TRACKER_FLAVOR
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.mlflow.experiment_trackers import (
        MLFlowExperimentTracker,
    )


def is_remote_mlflow_tracking_uri(tracking_uri: str) -> bool:
    """Checks whether the given tracking uri is remote or not.

    Args:
        tracking_uri: The tracking uri to check.

    Returns:
        `True` if the tracking uri is remote, `False` otherwise.
    """
    return any(
        tracking_uri.startswith(prefix) for prefix in ["http://", "https://"]
    ) or is_databricks_tracking_uri(tracking_uri)


def is_databricks_tracking_uri(tracking_uri: str) -> bool:
    """Checks whether the given tracking uri is a Databricks tracking uri.

    Args:
        tracking_uri: The tracking uri to check.

    Returns:
        `True` if the tracking uri is a Databricks tracking uri, `False`
        otherwise.
    """
    return tracking_uri == "databricks"


class MLFlowExperimentTrackerSettings(BaseSettings):
    """Settings for the MLflow experiment tracker."""

    experiment_name: Optional[str] = Field(
        None,
        description="The MLflow experiment name to use for tracking runs.",
    )
    nested: bool = Field(
        False,
        description="If `True`, will create a nested sub-run for the step.",
    )
    tags: Dict[str, Any] = Field(
        default_factory=dict,
        description="Tags to attach to the MLflow run for categorization and filtering.",
    )


class MLFlowExperimentTrackerConfig(
    BaseExperimentTrackerConfig, MLFlowExperimentTrackerSettings
):
    """Config for the MLflow experiment tracker."""

    tracking_uri: Optional[str] = Field(
        None,
        description="The URI of the MLflow tracking server. If no URI is set, "
        "your stack must contain a LocalArtifactStore and ZenML will point "
        "MLflow to a subdirectory of your artifact store instead.",
    )
    tracking_username: Optional[str] = SecretField(
        default=None,
        description="Username for authenticating with the MLflow tracking server. "
        "Required when using a remote tracking URI along with tracking_password.",
    )
    tracking_password: Optional[str] = SecretField(
        default=None,
        description="Password for authenticating with the MLflow tracking server. "
        "Required when using a remote tracking URI along with tracking_username.",
    )
    tracking_token: Optional[str] = SecretField(
        default=None,
        description="Token for authenticating with the MLflow tracking server. "
        "Alternative to username/password authentication for remote tracking URIs.",
    )
    tracking_insecure_tls: bool = Field(
        False,
        description="Skips verification of TLS connection to the MLflow tracking "
        "server if set to `True`. Use with caution in production environments.",
    )
    databricks_host: Optional[str] = Field(
        None,
        description="The host of the Databricks workspace with the MLflow managed "
        "server to connect to. Required when tracking_uri is set to 'databricks'.",
    )
    enable_unity_catalog: bool = Field(
        False,
        description="If `True`, will enable the Databricks Unity Catalog for "
        "logging and registering models.",
    )

    @model_validator(mode="after")
    def _ensure_authentication_if_necessary(
        self,
    ) -> "MLFlowExperimentTrackerConfig":
        """Ensures that credentials or a token for authentication exist.

        We make this check when running MLflow tracking with a remote backend.

        Returns:
            The validated values.

        Raises:
             ValueError: If neither credentials nor a token are provided.
        """
        if self.tracking_uri:
            if is_databricks_tracking_uri(self.tracking_uri):
                # If the tracking uri is "databricks", then we need the
                # databricks host to be set.
                if not self.databricks_host:
                    raise ValueError(
                        "MLflow experiment tracking with a Databricks MLflow "
                        "managed tracking server requires the "
                        "`databricks_host` to be set in your stack component. "
                        "To update your component, run "
                        "`zenml experiment-tracker update "
                        "<NAME> --databricks_host=DATABRICKS_HOST` "
                        "and specify the hostname of your Databricks workspace."
                    )

            if is_remote_mlflow_tracking_uri(self.tracking_uri):
                # we need either username + password or a token to authenticate
                # to the remote backend
                basic_auth = self.tracking_username and self.tracking_password

                if not (basic_auth or self.tracking_token):
                    raise ValueError(
                        f"MLflow experiment tracking with a remote backend "
                        f"{self.tracking_uri} is only possible when specifying "
                        f"either username and password or an authentication "
                        f"token in your stack component. To update your "
                        f"component, run the following command: "
                        f"`zenml experiment-tracker update "
                        f"<NAME> --tracking_username=MY_USERNAME "
                        f"--tracking_password=MY_PASSWORD "
                        f"--tracking_token=MY_TOKEN` and specify either your "
                        f"username and password or token."
                    )

        return self

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        if not self.tracking_uri or not is_remote_mlflow_tracking_uri(
            self.tracking_uri
        ):
            return True
        return False


class MLFlowExperimentTrackerFlavor(BaseExperimentTrackerFlavor):
    """Class for the `MLFlowExperimentTrackerFlavor`."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return MLFLOW_MODEL_EXPERIMENT_TRACKER_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the flavor.

        Returns:
            The display name of the flavor.
        """
        return "MLflow"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/experiment_tracker/mlflow.png"

    @property
    def config_class(self) -> Type[MLFlowExperimentTrackerConfig]:
        """Returns `MLFlowExperimentTrackerConfig` config class.

        Returns:
                The config class.
        """
        return MLFlowExperimentTrackerConfig

    @property
    def implementation_class(self) -> Type["MLFlowExperimentTracker"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.mlflow.experiment_trackers import (
            MLFlowExperimentTracker,
        )

        return MLFlowExperimentTracker



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base abstractions for the pub/sub layer."""

import asyncio
import logging
import random
from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Any

from pydantic import Field

from zenml.constants import (
    ENV_ZENML_CONSUMER_POLLING_PREFIX,
    ENV_ZENML_CONSUMER_PREFIX,
    ENV_ZENML_PRODUCER_PREFIX,
)
from zenml.pub_sub.models import (
    CriticalEvent,
    CriticalEventType,
    MessageEnvelope,
    MessagePayload,
)
from zenml.utils.env_utils import ConfigBase

logger = logging.getLogger(__name__)


class MessageDecodeError(Exception):
    """Raised when a raw broker message cannot be decoded/preprocessed."""


class MessageExecutionError(Exception):
    """Raised when message processing fails after retries."""


class MessageEncodeError(Exception):
    """Raised when a payload cannot be encoded into a broker message."""


class MessageSubmissionError(Exception):
    """Raised when sending a message fails."""


class ProducerConfig(ConfigBase):
    """Config object grouping producer configuration params."""

    @staticmethod
    def prefixes() -> list[str]:
        """Implementation of the prefixes config method.

        Returns:
            A list of prefixes to use for env var name resolution.
        """
        return [ENV_ZENML_PRODUCER_PREFIX]

    send_retries: int = Field(
        default=1,
        description="Number of retries to send messages before failing.",
        ge=0,
    )


class CriticalEventHandler(ABC):
    """Abstraction for handling critical events."""

    @abstractmethod
    async def handle(self, event: CriticalEvent) -> None:
        """Handles a critical event.

        Handling of a critical event would be include the following operations:
        - Tracking of the event instance and their evolution
        - Alerting mechanisms based on event monitoring alarms
        - Mitigation actions

        Args:
            event: A CriticalEvent instance.
        """
        pass


class ProducerBase(ABC):
    """Async producer base.

    Backends implement:
      1) build_message(payload) -> backend-specific message object (Any)
      2) send_message(message) -> submit to broker

    Base provides:
      - publish(payload): build + send (+ optional retries) and emits CriticalEvent on failure.
    """

    def __init__(
        self,
        config: ProducerConfig,
        event_handler: CriticalEventHandler | None = None,
    ) -> None:
        """Producer constructor.

        Args:
            config: The producer config object.
            event_handler: A CriticalEventHandler instance (optional).
        """
        self._cfg = config
        self._event_handler = event_handler

    @property
    def config(self) -> ProducerConfig:
        """Implements the 'config' property.

        Returns:
            The producer config object.
        """
        return self._cfg

    @abstractmethod
    async def build_message(self, payload: MessagePayload) -> Any:
        """Convert the message to a broker-specific payload.

        Args:
            payload: A MessagePayload object.

        Returns:
            A broker-specific message.
        """

    @abstractmethod
    async def send_message(self, message: Any) -> str:
        """Sends a message to the broker.

        Args:
            message: A broker-specific message.

        Returns:
            The message ID generated by the broker.
        """

    async def handle_critical_event(self, event: CriticalEvent) -> None:
        """Critical event handler.

        A critical event indicates that the application behavior has degraded.
        For instance, the inability to serialize the message to a broker-compatible format
        is extremely severe as it most probably repeats across messages meaning no messages
        are published for execution at all.

        A complete implementation of a critical event handling process would be the following:

        1. Log in detailed fashion the event so that it is traceable in the logs.
        2. Publish the event to an event aggregator service so that it is tracked long-term and
        alerts can be generated based on its frequency and severity.
        3. Hold the message in a DLQ to enable re-execution when a bug fix is introduced in the code.

        Args:
            event: A critical event instance.
        """
        logger.exception(
            "Critical event %s caused by the following exception:",
            event.value,
            exc_info=event.exception if event.exception else None,
        )
        if self._event_handler:
            asyncio.create_task(self._event_handler.handle(event))

    async def publish(self, payload: MessagePayload) -> str:
        """Build and send message, emitting CriticalEvent on failure.

        Args:
            payload: Application payload.

        Returns:
            Message ID generated by the broker.

        Raises:
            MessageEncodeError: If encoding/building fails.
            MessageSubmissionError: If sending fails after retries.
        """
        try:
            message = await self.build_message(payload)
        except Exception as exc:
            await self.handle_critical_event(
                CriticalEvent(
                    value=CriticalEventType.ENCODE_FAILED,
                    description=f"Encode failed for payload {payload.id}: {exc!r}",
                    created_at=datetime.now(tz=timezone.utc),
                    exception=exc,
                )
            )
            raise MessageEncodeError(
                f"Encode failed for payload {payload.id}"
            ) from exc

        attempts = self.config.send_retries + 1
        last_exc = None

        for attempt in range(attempts):
            try:
                return await self.send_message(message)
            except Exception as exc:
                last_exc = exc
                logger.exception(
                    "Attempt %s to send message %s failed",
                    attempt,
                    payload.id,
                )

        await self.handle_critical_event(
            CriticalEvent(
                value=CriticalEventType.SEND_FAILED,
                description=f"Submission failed for payload {payload.id}",
                created_at=datetime.now(tz=timezone.utc),
                exception=last_exc,
            )
        )
        raise MessageSubmissionError(f"Send failed for payload {payload.id}")


class ConsumerRuntimeConfig(ConfigBase):
    """Config shared by polling and push consumers."""

    @staticmethod
    def prefixes() -> list[str]:
        """Implementation of the prefixes config method.

        Returns:
            A list of prefixes to use for env var name resolution.
        """
        return [ENV_ZENML_CONSUMER_PREFIX]

    late_ack: bool = Field(
        default=True,
        description="Flag specifying acknowledgment pattern. If set to True "
        "will ack message after execution. Else if will ack the "
        "message before execution.",
    )
    execution_retries: int = Field(
        default=0,
        description="Number of execution retries.",
        ge=0,
    )
    ack_retries: int = Field(
        default=3,
        description="Number of times consumer will retry to ack message.",
        ge=0,
    )
    ack_retry_backoff: int = Field(
        default=1,
        description="Number of seconds between ack message retries.",
        ge=1,
    )
    process_retry_backoff: int = Field(
        default=1,
        description="Number of seconds between process message retries.",
        ge=1,
    )


class Executor(ABC):
    """Abstract executor interface."""

    @abstractmethod
    async def execute(self, payload: MessagePayload) -> None:
        """Executor main method.

        Responsible with executing the payload and adding implementing
        additional layers of error handling and retries:

        - Should decide whether to propagate errors to the queue layer.
        - Should decide whether retries are applicable.

        Args:
            payload: A standard message payload.
        """

    @abstractmethod
    async def shutdown(self) -> None:
        """Graceful executor shutdown (completes work, then exits)."""
        pass

    @abstractmethod
    def max_message_capacity(self) -> int | None:
        """Helper utility for variable batch size consumption.

        Helpful when execution needs to be throttled. The executor
        responds with the number of messages it can pickup for execution
        at the moment.

        Returns:
            Number of messages the executor can handle (None if no limit).
        """


class ConsumerBase(ABC):
    """Async core pipeline shared by polling and push transports.

    The transport decides *how* raw messages arrive. The base class provides:
      - decode/preprocess
      - execute with retries
      - ack with retries (+ critical event on exhaustion)
      - invalid/failing handlers
    """

    def __init__(
        self,
        config: ConsumerRuntimeConfig,
        executor: Executor,
        event_handler: CriticalEventHandler | None = None,
    ) -> None:
        """Consumer base constructor.

        Args:
            config: The consumer config object.
            executor: An executor instance.
            event_handler: A CriticalEventHandler instance (optional).
        """
        self._cfg = config
        self._executor = executor
        self._event_handler = event_handler

    @property
    def config(self) -> ConsumerRuntimeConfig:
        """Implements the 'config' property.

        Returns:
            The consumer config object.
        """
        return self._cfg

    # -------- backend hooks (required) --------

    @abstractmethod
    async def preprocess_message(self, raw_message: Any) -> MessageEnvelope:
        """Decode/Validate broker message and wrap to MessageEnvelope.

        Args:
            raw_message: Broker-specific message.

        Returns:
            A MessageEnvelope object.
        """

    async def process_message(self, message: MessageEnvelope) -> None:
        """Execution of message payload.

        Args:
            message: A MessageEnvelope object.
        """
        await self._executor.execute(payload=message.payload)

    @abstractmethod
    async def acknowledge_message(self, raw_message: Any) -> None:
        """Acks/Deletes message from the queue.

        Args:
            raw_message: The raw broker-specific message.
        """

    @abstractmethod
    async def handle_invalid_message(self, raw_message: Any) -> None:
        """Handles message of invalid format (not suitable for execution).

        Messages that are invalid should still be acked to avoid
        looping execution and system poisoning.

        Args:
            raw_message: The raw broker-specific message.

        """

    async def handle_critical_event(self, event: CriticalEvent) -> None:
        """Critical event handler.

        A critical event indicates that the application behavior has degraded.
        For instance, the inability to serialize the message to a broker-compatible format
        is extremely severe as it most probably repeats across messages meaning no messages
        are published for execution at all.

        A complete implementation of a critical event handling process would be the following:

        1. Log in detailed fashion the event so that it is traceable in the logs.
        2. Publish the event to an event aggregator service so that it is tracked long-term and
        alerts can be generated based on its frequency and severity.
        3. Hold the message in a DLQ to enable re-execution when a bug fix is introduced in the code.

        Args:
            event: A critical event instance.
        """
        logger.exception(
            "Critical event %s caused by the following exception:",
            event.value,
            exc_info=event.exception,
        )
        if self._event_handler:
            asyncio.create_task(self._event_handler.handle(event))

    async def handle_raw_message(self, raw_message: Any) -> None:
        """Handles one message at a time.

        Args:
            raw_message: A raw broker-specific message.
        """
        try:
            msg = await self.preprocess_message(raw_message)
        except MessageDecodeError:
            await self._ack_with_retries(raw_message=raw_message)
            await self.handle_invalid_message(raw_message)
            return

        if self.config.late_ack:
            await self._process_with_retries(msg)
            await self._ack_with_retries(raw_message=msg.raw_message)
        else:
            await self._ack_with_retries(raw_message=msg.raw_message)
            await self._process_with_retries(msg)

    async def _process_with_retries(self, message: MessageEnvelope) -> None:
        attempts = self.config.execution_retries + 1
        last_exc = None

        for attempt in range(attempts):
            try:
                await self.process_message(message)
                return
            except Exception as exc:
                last_exc = exc
                await asyncio.sleep(self.config.process_retry_backoff)

        await self.handle_critical_event(
            CriticalEvent(
                value=CriticalEventType.EXECUTE_FAILED,
                description=f"Execution failed for message {message.id}",
                created_at=datetime.now(tz=timezone.utc),
                exception=last_exc,
            )
        )

    async def _ack_with_retries(self, raw_message: Any) -> None:
        attempts = self.config.ack_retries + 1

        last_exc = None

        for attempt in range(1, attempts + 1):
            try:
                await self.acknowledge_message(raw_message=raw_message)
                return
            except Exception as exc:
                last_exc = exc
                await asyncio.sleep(self.config.ack_retry_backoff)

        await self.handle_critical_event(
            CriticalEvent(
                value=CriticalEventType.ACK_FAILED,
                description="Ack failed for message",
                created_at=datetime.now(tz=timezone.utc),
                exception=last_exc,
            )
        )

    @abstractmethod
    async def run(self) -> None:
        """Starts the consumer task."""
        pass


class PollingConfig(ConsumerRuntimeConfig):
    """Base config class for polling consumers."""

    @staticmethod
    def prefixes() -> list[str]:
        """Implementation of the prefixes config method.

        Returns:
            A list of prefixes to use for env var name resolution.
        """
        return ConsumerRuntimeConfig.prefixes() + [
            ENV_ZENML_CONSUMER_POLLING_PREFIX
        ]

    interval: float = Field(
        default=10.0,
        description="Polling interval in seconds.",
        ge=0.0,
    )
    jitter: float = Field(
        default=0.0,
        description="Jitter(+- random diff) for interval in seconds.",
        ge=0.0,
    )


class PollingConsumer(ConsumerBase, ABC):
    """Polling transport: the consumer fetches batches periodically."""

    def __init__(
        self,
        config: PollingConfig,
        executor: Executor,
        event_handler: CriticalEventHandler | None = None,
    ) -> None:
        """Polling consumer constructor.

        Args:
            config: Polling config instance.
            executor: Executor instance.
            event_handler: A CriticalEventHandler instance (optional).
        """
        super().__init__(
            config, executor=executor, event_handler=event_handler
        )
        self._stopped: bool = False

        # Jitter is computed once at construction (fixed for this consumer instance).
        self._polling_interval = config.interval + random.uniform(
            -config.jitter, config.jitter
        )

    def stop(self) -> None:
        """Signal the run loop to stop."""
        self._stopped = True

    @property
    def polling_interval(self) -> float:
        """Implement the 'polling_interval' property.

        Returns:
            The polling interval value.
        """
        return self._polling_interval

    @abstractmethod
    async def receive_messages(
        self, batch_size: int | None = None
    ) -> list[Any]:
        """Requests to get a batch of messages from the queue.

        Args:
            batch_size: Number of messages to get.

        Returns:
            A list of messages (broker-specific payload).
        """

    @abstractmethod
    def get_messages_batch_size(self) -> int:
        """Getter for the batch size.

        Enables implementations that adapt batch size per iteration
        based on configuration, throttling or other factors.

        Returns:
            The size of the next batch of messages.
        """

    async def poll_once(self) -> None:
        """Receive and handle one batch. Receive failures emit read_failed."""
        try:
            batch_size = self.get_messages_batch_size()
            raw_messages = await self.receive_messages(batch_size=batch_size)
        except Exception as exc:
            await self.handle_critical_event(
                CriticalEvent(
                    value=CriticalEventType.READ_FAILED,
                    description=f"Receive failed: {exc!r}",
                    created_at=datetime.now(tz=timezone.utc),
                    exception=exc,
                )
            )
            return

        for raw in raw_messages:
            await self.handle_raw_message(raw)

    async def run(self) -> None:
        """Run polling loop until stop() is called."""
        try:
            while not self._stopped:
                try:
                    start = asyncio.get_running_loop().time()
                    await self.poll_once()

                    elapsed = asyncio.get_running_loop().time() - start

                    if elapsed > self._polling_interval:
                        logger.debug("Elapsed time for polling: %s", elapsed)
                    remaining = self.polling_interval - elapsed
                    if remaining > 0:
                        await asyncio.sleep(remaining)
                except Exception as exc:
                    logger.exception(
                        "Unhandled exception during polling: %s", exc_info=exc
                    )
        finally:
            self.stop()
            await self._executor.shutdown()



"""Release [0.94.1].

Revision ID: 0.94.1
Revises: 24771f985eac
Create Date: 2026-03-19 11:15:07.142707

"""

# revision identifiers, used by Alembic.
revision = "0.94.1"
down_revision = "24771f985eac"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



"""add labels for stack components [43a86093b60e].

Revision ID: 43a86093b60e
Revises: 0.37.0
Create Date: 2023-03-28 11:26:33.358600

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "43a86093b60e"
down_revision = "0.37.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("labels", sa.LargeBinary(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("stack_component", schema=None) as batch_op:
        batch_op.drop_column("labels")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of pipeline deployments table."""

import json
from typing import TYPE_CHECKING, Any, List, Optional, Sequence
from uuid import UUID

from sqlalchemy import TEXT, Column, UniqueConstraint
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlalchemy.orm import selectinload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship, String

from zenml.constants import MEDIUMTEXT_MAX_LENGTH
from zenml.enums import (
    DeploymentStatus,
    TaggableResourceTypes,
    VisualizationResourceTypes,
)
from zenml.logger import get_logger
from zenml.models.v2.core.deployment import (
    DeploymentRequest,
    DeploymentResponse,
    DeploymentResponseBody,
    DeploymentResponseMetadata,
    DeploymentResponseResources,
    DeploymentUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.component_schemas import StackComponentSchema
from zenml.zen_stores.schemas.pipeline_snapshot_schemas import (
    PipelineSnapshotSchema,
)
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.curated_visualization_schemas import (
        CuratedVisualizationSchema,
    )
    from zenml.zen_stores.schemas.tag_schemas import TagSchema

logger = get_logger(__name__)


class DeploymentSchema(NamedSchema, table=True):
    """SQL Model for pipeline deployment."""

    __tablename__ = "deployment"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "project_id",
            name="unique_deployment_name_in_project",
        ),
    )

    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    project: "ProjectSchema" = Relationship(back_populates="deployments")

    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    user: Optional["UserSchema"] = Relationship(back_populates="deployments")

    status: str
    url: Optional[str] = Field(
        default=None,
        sa_column=Column(TEXT, nullable=True),
    )
    auth_key: Optional[str] = Field(
        default=None,
        sa_column=Column(TEXT, nullable=True),
    )
    deployment_metadata: str = Field(
        default="{}",
        sa_column=Column(
            String(length=MEDIUMTEXT_MAX_LENGTH).with_variant(
                MEDIUMTEXT, "mysql"
            ),
            nullable=False,
        ),
    )
    snapshot_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineSnapshotSchema.__tablename__,
        source_column="snapshot_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    snapshot: Optional["PipelineSnapshotSchema"] = Relationship(
        back_populates="deployment",
    )

    deployer_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=StackComponentSchema.__tablename__,
        source_column="deployer_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    deployer: Optional["StackComponentSchema"] = Relationship()

    tags: List["TagSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=f"and_(foreign(TagResourceSchema.resource_type)=='{TaggableResourceTypes.DEPLOYMENT.value}', foreign(TagResourceSchema.resource_id)==DeploymentSchema.id)",
            secondary="tag_resource",
            secondaryjoin="TagSchema.id == foreign(TagResourceSchema.tag_id)",
            order_by="TagSchema.name",
            overlaps="tags",
        ),
    )

    visualizations: List["CuratedVisualizationSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            primaryjoin=(
                "and_(CuratedVisualizationSchema.resource_type"
                f"=='{VisualizationResourceTypes.DEPLOYMENT.value}', "
                "foreign(CuratedVisualizationSchema.resource_id)==DeploymentSchema.id)"
            ),
            overlaps="visualizations",
            cascade="delete",
            order_by="CuratedVisualizationSchema.display_order",
        ),
    )

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    selectinload(jl_arg(DeploymentSchema.user)),
                    selectinload(jl_arg(DeploymentSchema.deployer)),
                    selectinload(jl_arg(DeploymentSchema.snapshot)).joinedload(
                        jl_arg(PipelineSnapshotSchema.pipeline)
                    ),
                    selectinload(jl_arg(DeploymentSchema.snapshot)).joinedload(
                        jl_arg(PipelineSnapshotSchema.stack)
                    ),
                    selectinload(jl_arg(DeploymentSchema.visualizations)),
                ]
            )

        return options

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> DeploymentResponse:
        """Convert a `DeploymentSchema` to a `DeploymentResponse`.

        Args:
            include_metadata: Whether to include metadata in the response.
            include_resources: Whether to include resources in the response.
            kwargs: Additional keyword arguments.

        Returns:
            The created `DeploymentResponse`.
        """
        status: Optional[DeploymentStatus] = None
        if self.status in DeploymentStatus.values():
            status = DeploymentStatus(self.status)
        elif self.status is not None:
            status = DeploymentStatus.UNKNOWN
            logger.warning(
                f"Deployment status '{self.status}' used for deployment "
                f"{self.name} is not a valid DeploymentStatus value. "
                "Using UNKNOWN instead."
            )

        body = DeploymentResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            created=self.created,
            updated=self.updated,
            url=self.url,
            status=status,
        )

        metadata = None
        if include_metadata:
            metadata = DeploymentResponseMetadata(
                deployment_metadata=json.loads(self.deployment_metadata),
                auth_key=self.auth_key,
            )

        resources = None
        if include_resources:
            resources = DeploymentResponseResources(
                user=self.user.to_model() if self.user else None,
                tags=[tag.to_model() for tag in self.tags],
                snapshot=self.snapshot.to_model() if self.snapshot else None,
                deployer=self.deployer.to_model() if self.deployer else None,
                pipeline=self.snapshot.pipeline.to_model()
                if self.snapshot and self.snapshot.pipeline
                else None,
                stack=self.snapshot.stack.to_model()
                if self.snapshot and self.snapshot.stack
                else None,
                visualizations=[
                    visualization.to_model(
                        include_metadata=False,
                        include_resources=False,
                    )
                    for visualization in self.visualizations
                ],
            )

        return DeploymentResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(
        self,
        update: DeploymentUpdate,
    ) -> "DeploymentSchema":
        """Updates a `DeploymentSchema` from a `DeploymentUpdate`.

        Args:
            update: The `DeploymentUpdate` to update from.

        Returns:
            The updated `DeploymentSchema`.
        """
        for field, value in update.model_dump(
            exclude_unset=True, exclude_none=True
        ).items():
            if field == "deployment_metadata":
                setattr(self, field, json.dumps(value))
            elif hasattr(self, field):
                setattr(self, field, value)

        self.updated = utc_now()
        return self

    @classmethod
    def from_request(cls, request: DeploymentRequest) -> "DeploymentSchema":
        """Convert a `DeploymentRequest` to a `DeploymentSchema`.

        Args:
            request: The request model to convert.

        Returns:
            The converted schema.
        """
        return cls(
            name=request.name,
            project_id=request.project,
            user_id=request.user,
            status=DeploymentStatus.UNKNOWN.value,
            snapshot_id=request.snapshot_id,
            deployer_id=request.deployer_id,
            auth_key=request.auth_key,
        )



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Entrypoint of the Lightning master/orchestrator STUDIO."""

import argparse
import os
from typing import Dict, cast

from lightning_sdk import Machine, Studio

from zenml.client import Client
from zenml.entrypoints.step_entrypoint_configuration import (
    StepEntrypointConfiguration,
)
from zenml.enums import ExecutionStatus
from zenml.integrations.lightning.flavors.lightning_orchestrator_flavor import (
    LightningOrchestratorSettings,
)
from zenml.integrations.lightning.orchestrators.lightning_orchestrator import (
    ENV_ZENML_LIGHTNING_ORCHESTRATOR_RUN_ID,
    LightningOrchestrator,
)
from zenml.integrations.lightning.orchestrators.utils import (
    gather_requirements,
    sanitize_studio_name,
)
from zenml.logger import get_logger
from zenml.orchestrators.dag_runner import ThreadedDagRunner

logger = get_logger(__name__)


LIGHTNING_ZENML_DEFAULT_CUSTOM_REPOSITORY_PATH = "."


def parse_args() -> argparse.Namespace:
    """Parse entrypoint arguments.

    Returns:
        Parsed args.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("--run_name", type=str, required=True)
    parser.add_argument("--snapshot_id", type=str, required=True)
    return parser.parse_args()


def main() -> None:
    """Entrypoint of the Lightning master/orchestrator STUDIO.

    This is the entrypoint of the Lightning master/orchestrator STUDIO. It is
    responsible for provisioning the STUDIO and running the pipeline steps in
    separate STUDIO.

    Raises:
        TypeError: If the active stack's orchestrator is not an instance of
            LightningOrchestrator.
        ValueError: If the active stack's container registry is None.
    """
    # Log to the container's stdout so it can be streamed by the client.
    logger.info("Lightning orchestrator STUDIO started.")

    # Parse / extract args.
    args = parse_args()

    orchestrator_run_id = os.environ.get(
        ENV_ZENML_LIGHTNING_ORCHESTRATOR_RUN_ID
    )
    if not orchestrator_run_id:
        raise ValueError(
            f"Environment variable '{ENV_ZENML_LIGHTNING_ORCHESTRATOR_RUN_ID}' is not set."
        )

    logger.info(f"Orchestrator run id: {orchestrator_run_id}")

    snapshot = Client().get_snapshot(args.snapshot_id)
    filename = f"{args.run_name}.tar.gz"

    pipeline_dag = {
        step_name: step.spec.upstream_steps
        for step_name, step in snapshot.step_configurations.items()
    }
    entrypoint_command = StepEntrypointConfiguration.get_entrypoint_command()

    active_stack = Client().active_stack

    orchestrator = active_stack.orchestrator
    if not isinstance(orchestrator, LightningOrchestrator):
        raise TypeError(
            "The active stack's orchestrator is not an instance of LightningOrchestrator."
        )

    # Set up credentials
    orchestrator._set_lightning_env_vars(snapshot)

    pipeline_settings = cast(
        LightningOrchestratorSettings, orchestrator.get_settings(snapshot)
    )

    # Gather the requirements
    pipeline_docker_settings = snapshot.pipeline_configuration.docker_settings
    pipeline_requirements = gather_requirements(pipeline_docker_settings)
    pipeline_requirements_to_string = " ".join(
        f'"{req}"' for req in pipeline_requirements
    )

    unique_resource_configs: Dict[str, str] = {}
    main_studio_name = sanitize_studio_name(
        f"zenml_{orchestrator_run_id}_pipeline"
    )
    for step_name, step in snapshot.step_configurations.items():
        step_settings = cast(
            LightningOrchestratorSettings,
            orchestrator.get_settings(step),
        )
        unique_resource_configs[step_name] = main_studio_name
        if pipeline_settings.machine_type != step_settings.machine_type:
            unique_resource_configs[step_name] = (
                f"zenml-{orchestrator_run_id}_{step_name}"
            )

    logger.info(f"Creating main studio: {main_studio_name}")
    main_studio = Studio(name=main_studio_name)
    if pipeline_settings.machine_type:
        main_studio.start(Machine(pipeline_settings.machine_type))
    else:
        main_studio.start()

    logger.info("Main studio started.")
    logger.info("Uploading code to main studio the code path: %s", filename)
    main_studio.upload_file(
        "/teamspace/studios/this_studio/.lightning_studio/.studiorc",
        remote_path="/teamspace/studios/this_studio/.lightning_studio/.studiorc",
    )
    output = main_studio.run(
        f"mkdir -p /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]}"
    )
    logger.info(output)
    main_studio.upload_file(
        f"/teamspace/studios/this_studio/zenml_codes/{filename}",
        remote_path=f"/teamspace/studios/this_studio/zenml_codes/{filename}",
    )
    logger.info("Extracting code... ")
    output = main_studio.run(
        f"tar -xvzf /teamspace/studios/this_studio/zenml_codes/{filename} -C /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]}"
    )
    logger.info(f"Code extraction output: {output}")
    logger.info("Installing requirements... ")

    output = main_studio.run("pip install uv")
    logger.info(output)
    output = main_studio.run(
        f"uv pip install {pipeline_requirements_to_string}"
    )
    logger.info(output)
    output = main_studio.run("pip install zenml")
    logger.info(output)

    for command in pipeline_settings.custom_commands or []:
        output = main_studio.run(
            f"cd /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]} && {command}"
        )
        logger.info(f"Custom command output: {output}")

    run = Client().list_pipeline_runs(
        sort_by="asc:created",
        size=1,
        snapshot_id=args.snapshot_id,
        status=ExecutionStatus.INITIALIZING,
    )[0]

    logger.info("Fetching pipeline run: %s", run.id)

    def run_step_on_lightning_studio(step_name: str) -> None:
        """Run a pipeline step in a separate Lightning STUDIO.

        Args:
            step_name: Name of the step.

        Raises:
            Exception: If an error occurs while running the step on the STUDIO.
        """
        step_args = StepEntrypointConfiguration.get_entrypoint_arguments(
            step_name=step_name,
            snapshot_id=args.snapshot_id,
        )

        entrypoint = entrypoint_command + step_args
        entrypoint_string = " ".join(entrypoint)
        run_command = f"{entrypoint_string}"

        step = snapshot.step_configurations[step_name]
        if unique_resource_configs[step_name] != main_studio_name:
            logger.info(
                f"Creating separate studio for step: {unique_resource_configs[step_name]}"
            )
            # Get step settings
            step_settings = cast(
                LightningOrchestratorSettings,
                orchestrator.get_settings(step),
            )
            # Gather the requirements
            step_docker_settings = step.config.docker_settings
            step_requirements = gather_requirements(step_docker_settings)
            step_requirements_to_string = " ".join(
                f'"{req}"' for req in step_requirements
            )
            run_command = f"{entrypoint_string}"

            logger.info(
                f"Creating separate studio for step: {unique_resource_configs[step_name]}"
            )
            studio = Studio(name=unique_resource_configs[step_name])
            try:
                studio.start(Machine(step_settings.machine_type))
                output = studio.run(
                    f"mkdir -p /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]}"
                )
                logger.info(output)
                studio.upload_file(
                    f"/teamspace/studios/this_studio/zenml_codes/{filename}",
                    remote_path=f"/teamspace/studios/this_studio/zenml_codes/{filename}",
                )
                output = studio.run(
                    f"tar -xvzf /teamspace/studios/this_studio/zenml_codes/{filename} -C /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]}"
                )
                logger.info(output)
                studio.upload_file(
                    "/teamspace/studios/this_studio/.lightning_studio/.studiorc",
                    remote_path="/teamspace/studios/this_studio/.lightning_studio/.studiorc",
                )
                output = studio.run("pip install uv")
                logger.info(output)
                output = studio.run(
                    f"uv pip install {step_requirements_to_string}"
                )
                logger.info(output)
                output = studio.run("pip install zenml")
                logger.info(output)
                for command in step_settings.custom_commands or []:
                    output = studio.run(
                        f"cd /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]} && {command}"
                    )
                    logger.info(f"Custom command output: {output}")
                output = studio.run(
                    f"cd /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]} && {run_command}"
                )
                logger.info(output)
            except Exception as e:
                logger.error(
                    f"Error running step {step_name} on studio {unique_resource_configs[step_name]}: {e}"
                )
                raise e
            finally:
                studio.delete()
                studio.delete()
        else:
            output = main_studio.run(
                f"cd /teamspace/studios/this_studio/zenml_codes/{filename.rsplit('.', 2)[0]} && {run_command}"
            )
            logger.info(output)

            # Pop the resource configuration for this step
        unique_resource_configs.pop(step_name)

        if main_studio_name in unique_resource_configs.values():
            # If there are more steps using this configuration, skip deprovisioning the cluster
            logger.info(
                f"Resource configuration for studio '{main_studio_name}' "
                "is used by subsequent steps. Skipping the deprovisioning of "
                "the studio."
            )
        else:
            # If there are no more steps using this configuration, down the cluster
            logger.info(
                f"Resource configuration for cluster '{main_studio_name}' "
                "is not used by subsequent steps. deprovisioning the cluster."
            )
            main_studio.delete()
        logger.info(f"Running step `{step_name}` on a Studio is completed.")

    ThreadedDagRunner(
        dag=pipeline_dag, run_fn=run_step_on_lightning_studio
    ).run()

    logger.info("Orchestration STUDIO provisioned.")


if __name__ == "__main__":
    main()



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""OpenTelemetry exporter that writes logs to ZenML artifact store."""

import os
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Sequence
from uuid import uuid4

from opentelemetry.sdk._logs.export import LogExporter, LogExportResult

from zenml.artifact_stores.base_artifact_store import BaseArtifactStore
from zenml.enums import LoggingLevels
from zenml.log_stores.artifact.artifact_log_store import (
    END_OF_STREAM_MESSAGE,
    remove_ansi_escape_codes,
)
from zenml.logger import get_logger
from zenml.utils.logging_utils import LogEntry
from zenml.utils.time_utils import utc_now

if TYPE_CHECKING:
    from opentelemetry.sdk._logs import LogData


DEFAULT_MESSAGE_SIZE = 5 * 1024
LOGS_EXTENSION = ".log"

logger = get_logger(__name__)


class ArtifactLogExporter(LogExporter):
    """OpenTelemetry exporter that writes logs to ZenML artifact store."""

    def __init__(self, artifact_store: "BaseArtifactStore") -> None:
        """Initialize the exporter.

        Args:
            artifact_store: The artifact store to write logs to.
        """
        self.artifact_store = artifact_store

    def export(self, batch: Sequence["LogData"]) -> LogExportResult:
        """Export a batch of logs to the artifact store.

        Args:
            batch: Sequence of LogData to export (can be from multiple contexts).

        Returns:
            LogExportResult indicating success or failure.
        """
        if not batch:
            return LogExportResult.SUCCESS

        try:
            logs_by_uri: Dict[str, List[str]] = defaultdict(list)
            finalized_log_streams: List[str] = []

            for log_data in batch:
                attrs = log_data.log_record.attributes
                if not attrs:
                    continue

                log_uri = attrs.get("zenml.log.uri")
                if not log_uri or not isinstance(log_uri, str):
                    continue

                if log_data.log_record.body is END_OF_STREAM_MESSAGE:
                    finalized_log_streams.append(log_uri)
                    continue

                entries = self._otel_record_to_log_entries(log_data)
                for entry in entries:
                    json_line = entry.model_dump_json(exclude_none=True)
                    logs_by_uri[log_uri].append(json_line)

            for log_uri, log_lines in logs_by_uri.items():
                if log_lines:
                    self._write(log_lines, log_uri)

            for log_uri in finalized_log_streams:
                self._finalize(log_uri)

            return LogExportResult.SUCCESS

        except Exception:
            logger.exception("Failed to export logs to artifact store")
            return LogExportResult.FAILURE

    def _otel_record_to_log_entries(
        self, log_data: "LogData"
    ) -> List[LogEntry]:
        """Convert an OTEL log record to ZenML LogEntry objects.

        Args:
            log_data: The OpenTelemetry log data.

        Returns:
            List of LogEntry objects (multiple if message was chunked).
        """
        log_record = log_data.log_record
        attributes = log_record.attributes

        message = str(log_record.body) if log_record.body else ""
        if attributes and attributes.get("exception.message"):
            exc_message = attributes.get("exception.message")
            exc_type = attributes.get("exception.type")
            exc_stacktrace = attributes.get("exception.stacktrace")
            message += f"\n{exc_type}: {exc_message}\n{exc_stacktrace}"

        message = remove_ansi_escape_codes(message).rstrip()

        level = (
            self._map_severity_to_level(log_record.severity_text)
            if log_record.severity_text
            else None
        )

        name = "unknown"
        module = None
        filename = None
        lineno = None

        if log_record.attributes:
            attrs = dict(log_record.attributes)
            filename = attrs.get("code.filepath", None)
            lineno = attrs.get("code.lineno", None)
            module = attrs.get("code.function", None)

        message_bytes = message.encode("utf-8")
        if len(message_bytes) <= DEFAULT_MESSAGE_SIZE:
            return [
                LogEntry(
                    message=message,
                    name=name,
                    level=level,
                    timestamp=utc_now(tz_aware=True),
                    module=module,
                    filename=filename,
                    lineno=lineno,
                )
            ]
        else:
            chunks = self._split_to_chunks(message)
            entry_id = uuid4()
            entries = []

            for i, chunk in enumerate(chunks):
                entries.append(
                    LogEntry(
                        message=chunk,
                        name=name,
                        level=level,
                        timestamp=utc_now(tz_aware=True),
                        module=module,
                        filename=filename,
                        lineno=lineno,
                        chunk_index=i,
                        total_chunks=len(chunks),
                        id=entry_id,
                    )
                )

            return entries

    def _map_severity_to_level(self, severity_text: str) -> LoggingLevels:
        """Map OTEL severity text to ZenML LoggingLevels enum.

        Args:
            severity_text: The OTEL severity text.

        Returns:
            The corresponding LoggingLevels enum value.
        """
        if not severity_text:
            return LoggingLevels.INFO

        severity_upper = severity_text.upper()

        if severity_upper in ["DEBUG", "TRACE"]:
            return LoggingLevels.DEBUG
        elif severity_upper in ["INFO", "INFORMATION"]:
            return LoggingLevels.INFO
        elif severity_upper in ["WARN", "WARNING"]:
            return LoggingLevels.WARN
        elif severity_upper == "ERROR":
            return LoggingLevels.ERROR
        elif severity_upper in ["CRITICAL", "FATAL", "EMERGENCY"]:
            return LoggingLevels.CRITICAL
        else:
            return LoggingLevels.INFO

    def _split_to_chunks(self, message: str) -> List[str]:
        """Split a large message into chunks, handling UTF-8 boundaries.

        Args:
            message: The message to split.

        Returns:
            A list of message chunks.
        """
        message_bytes = message.encode("utf-8")
        chunks = []
        start = 0

        while start < len(message_bytes):
            end = min(start + DEFAULT_MESSAGE_SIZE, len(message_bytes))

            while end > start:
                chunk_bytes = message_bytes[start:end]
                try:
                    chunk_text = chunk_bytes.decode("utf-8")
                    chunks.append(chunk_text)
                    break
                except UnicodeDecodeError:
                    end -= 1
            else:
                end = min(start + DEFAULT_MESSAGE_SIZE, len(message_bytes))
                chunks.append(
                    message_bytes[start:end].decode("utf-8", errors="replace")
                )

            start = end

        return chunks

    def _write(
        self,
        log_lines: List[str],
        log_uri: str,
    ) -> None:
        """Write log lines to the artifact store.

        Args:
            log_lines: List of JSON-serialized log entries.
            log_uri: The URI of the log files to write.

        Raises:
            Exception: If the log lines cannot be written to the artifact store.
        """
        try:
            content = "\n".join(log_lines) + "\n"

            if self.artifact_store.config.IS_IMMUTABLE_FILESYSTEM:
                if not self.artifact_store.exists(log_uri):
                    self.artifact_store.makedirs(log_uri)

                timestamp = time.time()
                file_uri = os.path.join(
                    log_uri,
                    f"{timestamp}{LOGS_EXTENSION}",
                )

                with self.artifact_store.open(file_uri, "w") as f:
                    f.write(content)
            else:
                logs_base_uri = os.path.dirname(log_uri)
                if not self.artifact_store.exists(logs_base_uri):
                    self.artifact_store.makedirs(logs_base_uri)

                with self.artifact_store.open(log_uri, "a") as f:
                    f.write(content)

        except Exception as e:
            logger.error(f"Failed to write logs to {log_uri}: {e}")
            raise

    def _finalize(
        self,
        log_uri: str,
    ) -> None:
        """Finalize the logs for a given log model by merging all log files into one.

        Args:
            log_uri: The URI of the log files to finalize.

        Raises:
            Exception: If the logs cannot be finalized.
        """
        try:
            if self.artifact_store.config.IS_IMMUTABLE_FILESYSTEM:
                self._merge(log_uri)
            else:
                self.artifact_store._remove_previous_file_versions(log_uri)

        except Exception as e:
            logger.error(f"Failed to finalize logs for {log_uri}: {e}")
            raise

    def _merge(self, log_uri: str) -> None:
        """Merges all log files into one in the given URI.

        Called on the logging context exit.

        Args:
            log_uri: The URI of the log files to merge.
        """
        from zenml.artifacts.utils import _load_file_from_artifact_store
        from zenml.exceptions import DoesNotExistException

        # Check if the log directory exists - it may not if no logs
        # were written yet. The URI folder gets created only when the
        # first log message is sent.
        if not self.artifact_store.exists(log_uri):
            return

        files_ = self.artifact_store.listdir(log_uri)
        if len(files_) > 1:
            files_.sort()

            missing_files = set()
            # dump all logs to a local file first
            with self.artifact_store.open(
                os.path.join(log_uri, f"{time.time()}_merged{LOGS_EXTENSION}"),
                "w",
            ) as merged_file:
                for file in files_:
                    try:
                        merged_file.write(
                            str(
                                _load_file_from_artifact_store(
                                    os.path.join(log_uri, str(file)),
                                    artifact_store=self.artifact_store,
                                    mode="r",
                                )
                            )
                        )
                    except DoesNotExistException:
                        missing_files.add(file)

            # clean up left over files
            for file in files_:
                if file not in missing_files:
                    self.artifact_store.remove(
                        os.path.join(log_uri, str(file))
                    )

    def shutdown(self) -> None:
        """Shutdown the exporter."""
        self.artifact_store.cleanup()



"""Add user default project [1f1105204aaf].

Revision ID: 1f1105204aaf
Revises: 0.80.0
Create Date: 2025-03-25 11:11:13.480745

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "1f1105204aaf"
down_revision = "0.80.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "default_project_id",
                sqlmodel.sql.sqltypes.GUID(),
                nullable=True,
            )
        )
        batch_op.create_foreign_key(
            "fk_user_default_project_id_project",
            "project",
            ["default_project_id"],
            ["id"],
            ondelete="SET NULL",
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("user", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_user_default_project_id_project", type_="foreignkey"
        )
        batch_op.drop_column("default_project_id")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2026. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""PRO native schedules utility functions."""

import math
from datetime import datetime, timedelta, timezone

from croniter import croniter


def next_occurrence_for_interval(
    interval: int | float, start: datetime, base: datetime | None = None
) -> datetime:
    """Calculate the next occurrence based on interval based schedules.

    Args:
        interval: The difference (in seconds) between two occurrences.
        start: The first occurrence.
        base: The base used to calculate the next occurrence.

    Returns:
        The next occurrence datetime.
    """
    start = start.replace(tzinfo=timezone.utc)

    if not base:
        base = datetime.now(tz=timezone.utc).replace(second=0, microsecond=0)
    else:
        base = base.replace(tzinfo=timezone.utc)

    in_between_intervals = math.floor(
        (base - start).total_seconds() / interval
    )

    return (
        start + timedelta(seconds=(in_between_intervals + 1) * interval)
    ).replace(second=0, microsecond=0)


def next_occurrence_for_cron(
    expression: str, base: datetime | None = None
) -> datetime:
    """Calculate the next occurrence based on interval based schedules.

    Args:
        expression: The cron expression.
        base: The base used to calculate the next occurrence.

    Returns:
        The next occurrence datetime.
    """
    if not base:
        base = datetime.now(tz=timezone.utc)
    else:
        base = base.replace(tzinfo=timezone.utc)

    return croniter(expression, base).get_next(datetime)


def calculate_first_occurrence(
    cron_expression: str | None = None,
    interval: int | None = None,
    start_time: datetime | None = None,
    run_once_start_time: datetime | None = None,
) -> datetime | None:
    """Calculate the first occurrence of a schedule.

    Args:
        cron_expression: The cron expression.
        interval: The interval in seconds.
        start_time: The start time.
        run_once_start_time: The run once start time.

    Returns:
        The first occurrence of a schedule.
    """
    now_plus_1_sec = datetime.now(timezone.utc).replace(
        tzinfo=None
    ) + timedelta(seconds=1)

    if cron_expression:
        if start_time and start_time > now_plus_1_sec:
            base = start_time
        else:
            base = now_plus_1_sec

        next_occurrence = next_occurrence_for_cron(
            expression=cron_expression,
            base=base,
        )
    elif start_time and interval:
        if start_time > now_plus_1_sec:
            next_occurrence = start_time
        else:
            next_occurrence = next_occurrence_for_interval(
                interval=interval,
                start=start_time,
                base=now_plus_1_sec,
            )
    elif run_once_start_time:
        next_occurrence = run_once_start_time
    else:
        return None

    return next_occurrence



"""add flavor display_name [bc6317491676].

Revision ID: bc6317491676
Revises: 0.91.2
Create Date: 2025-11-20 16:00:00.000000

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op

# revision identifiers, used by Alembic.
revision = "bc6317491676"
down_revision = "0.91.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    with op.batch_alter_table("flavor", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "display_name",
                sqlmodel.sql.sqltypes.AutoString(),
                nullable=True,
            )
        )


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    with op.batch_alter_table("flavor", schema=None) as batch_op:
        batch_op.drop_column("display_name")



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Functionality to run ZenML steps or pipelines."""

import argparse

from zenml.entrypoints.base_entrypoint_configuration import (
    ENTRYPOINT_CONFIG_SOURCE_OPTION,
    BaseEntrypointConfiguration,
)
from zenml.execution.pipeline.utils import prevent_pipeline_execution
from zenml.utils import source_utils


def main() -> None:
    """Runs the entrypoint configuration given by the command line arguments."""
    # Make sure this entrypoint does not run an entire pipeline when
    # importing user modules. This could happen if a pipeline is called in a
    # module without an `if __name__== "__main__":` check)
    with prevent_pipeline_execution():
        parser = argparse.ArgumentParser()
        parser.add_argument(
            f"--{ENTRYPOINT_CONFIG_SOURCE_OPTION}", required=True
        )
        args, remaining_args = parser.parse_known_args()

        entrypoint_config_class = source_utils.load_and_validate_class(
            args.entrypoint_config_source,
            expected_class=BaseEntrypointConfiguration,
        )
        entrypoint_config = entrypoint_config_class(arguments=remaining_args)

        entrypoint_config.run()


if __name__ == "__main__":
    main()



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation for authorized OAuth2 devices."""

from datetime import datetime, timedelta
from secrets import token_hex
from typing import Any, Optional, Sequence, Tuple
from uuid import UUID

from passlib.context import CryptContext
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Relationship

from zenml.enums import OAuthDeviceStatus
from zenml.models import (
    OAuthDeviceInternalRequest,
    OAuthDeviceInternalResponse,
    OAuthDeviceInternalUpdate,
    OAuthDeviceResponse,
    OAuthDeviceResponseBody,
    OAuthDeviceResponseMetadata,
    OAuthDeviceResponseResources,
    OAuthDeviceUpdate,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import BaseSchema
from zenml.zen_stores.schemas.schema_utils import build_foreign_key_field
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import jl_arg


class OAuthDeviceSchema(BaseSchema, table=True):
    """SQL Model for authorized OAuth2 devices."""

    __tablename__ = "auth_devices"

    client_id: UUID
    user_code: str
    device_code: str
    status: str
    failed_auth_attempts: int = 0
    expires: Optional[datetime] = None
    last_login: Optional[datetime] = None
    trusted_device: bool = False
    os: Optional[str] = None
    ip_address: Optional[str] = None
    hostname: Optional[str] = None
    python_version: Optional[str] = None
    zenml_version: Optional[str] = None
    city: Optional[str] = None
    region: Optional[str] = None
    country: Optional[str] = None

    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    user: Optional["UserSchema"] = Relationship(back_populates="auth_devices")

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        options = []

        if include_resources:
            options.extend(
                [
                    joinedload(jl_arg(OAuthDeviceSchema.user)),
                ]
            )

        return options

    @classmethod
    def _generate_user_code(cls) -> str:
        """Generate a user code for an OAuth2 device.

        Returns:
            The generated user code.
        """
        return token_hex(16)

    @classmethod
    def _generate_device_code(cls) -> str:
        """Generate a device code.

        Returns:
            The generated device code.
        """
        return token_hex(32)

    @classmethod
    def _get_hashed_code(cls, code: str) -> str:
        """Hashes the input code and returns the hash value.

        Args:
            code: The code value to hash.

        Returns:
            The code hash value.
        """
        context = CryptContext(schemes=["bcrypt"], deprecated="auto")
        return context.hash(code)

    @classmethod
    def from_request(
        cls, request: OAuthDeviceInternalRequest
    ) -> Tuple["OAuthDeviceSchema", str, str]:
        """Create an authorized device DB entry from a device authorization request.

        Args:
            request: The device authorization request.

        Returns:
            The created `OAuthDeviceSchema`, the user code and the device code.
        """
        user_code = cls._generate_user_code()
        device_code = cls._generate_device_code()
        hashed_user_code = cls._get_hashed_code(user_code)
        hashed_device_code = cls._get_hashed_code(device_code)
        now = utc_now()
        return (
            cls(
                client_id=request.client_id,
                user_code=hashed_user_code,
                device_code=hashed_device_code,
                status=OAuthDeviceStatus.PENDING.value,
                failed_auth_attempts=0,
                expires=now + timedelta(seconds=request.expires_in),
                os=request.os,
                ip_address=request.ip_address,
                hostname=request.hostname,
                python_version=request.python_version,
                zenml_version=request.zenml_version,
                city=request.city,
                region=request.region,
                country=request.country,
                created=now,
                updated=now,
            ),
            user_code,
            device_code,
        )

    def update(self, device_update: OAuthDeviceUpdate) -> "OAuthDeviceSchema":
        """Update an authorized device from a device update model.

        Args:
            device_update: The device update model.

        Returns:
            The updated `OAuthDeviceSchema`.
        """
        for field, value in device_update.model_dump(
            exclude_none=True
        ).items():
            if hasattr(self, field):
                setattr(self, field, value)

        if device_update.locked is True:
            self.status = OAuthDeviceStatus.LOCKED.value
        elif device_update.locked is False:
            self.status = OAuthDeviceStatus.ACTIVE.value

        self.updated = utc_now()
        return self

    def internal_update(
        self, device_update: OAuthDeviceInternalUpdate
    ) -> Tuple["OAuthDeviceSchema", Optional[str], Optional[str]]:
        """Update an authorized device from an internal device update model.

        Args:
            device_update: The internal device update model.

        Returns:
            The updated `OAuthDeviceSchema` and the new user code and device
            code, if they were generated.
        """
        now = utc_now()
        user_code: Optional[str] = None
        device_code: Optional[str] = None

        # This call also takes care of setting fields that have the same
        # name in the internal model and the schema.
        self.update(device_update)

        if device_update.expires_in is not None:
            if device_update.expires_in <= 0:
                self.expires = None
            else:
                self.expires = now + timedelta(
                    seconds=device_update.expires_in
                )
        if device_update.update_last_login:
            self.last_login = now
        if device_update.generate_new_codes:
            user_code = self._generate_user_code()
            device_code = self._generate_device_code()
            self.user_code = self._get_hashed_code(user_code)
            self.device_code = self._get_hashed_code(device_code)
        self.updated = now
        return self, user_code, device_code

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> OAuthDeviceResponse:
        """Convert a device schema to a device response model.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The converted device response model.
        """
        metadata = None
        if include_metadata:
            metadata = OAuthDeviceResponseMetadata(
                python_version=self.python_version,
                zenml_version=self.zenml_version,
                city=self.city,
                region=self.region,
                country=self.country,
                failed_auth_attempts=self.failed_auth_attempts,
                last_login=self.last_login,
            )

        body = OAuthDeviceResponseBody(
            user_id=self.user_id,
            created=self.created,
            updated=self.updated,
            client_id=self.client_id,
            expires=self.expires,
            trusted_device=self.trusted_device,
            status=OAuthDeviceStatus(self.status),
            os=self.os,
            ip_address=self.ip_address,
            hostname=self.hostname,
        )
        resources = None
        if include_resources:
            resources = OAuthDeviceResponseResources(
                user=self.user.to_model() if self.user else None,
            )
        return OAuthDeviceResponse(
            id=self.id,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def to_internal_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
    ) -> OAuthDeviceInternalResponse:
        """Convert a device schema to an internal device response model.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.

        Returns:
            The converted internal device response model.
        """
        device_model = self.to_model(
            include_metadata=include_metadata,
            include_resources=include_resources,
        )
        return OAuthDeviceInternalResponse(
            id=device_model.id,
            body=device_model.body,
            metadata=device_model.metadata,
            resources=device_model.resources,
            user_code=self.user_code,
            device_code=self.device_code,
        )



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Warning configuration class and helper enums."""

from pydantic import BaseModel, Field

from zenml.utils.enum_utils import StrEnum


class WarningSeverity(StrEnum):
    """Enum class describing the warning severity."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class WarningCategory(StrEnum):
    """Enum class describing the warning category."""

    USAGE = "USAGE"


class WarningVerbosity(StrEnum):
    """Enum class describing the warning verbosity."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class WarningConfig(BaseModel, use_enum_values=True):
    """Warning config class describing how warning messages should be displayed."""

    description: str = Field(description="Description of the warning message")
    severity: WarningSeverity = WarningSeverity.MEDIUM
    category: WarningCategory
    code: str
    is_throttled: bool = Field(
        description="If the warning is throttled. Throttled warnings should be displayed once per session.",
        default=False,
    )
    verbosity: WarningVerbosity = Field(
        description="Verbosity of the warning. Low verbosity displays basic details (code & message). Medium displays also call details like module & line number. High displays description and additional info.",
        default=WarningVerbosity.MEDIUM,
    )



"""Release [0.56.4].

Revision ID: 0.56.4
Revises: 2d201872e23c
Create Date: 2024-04-24 08:05:30.664289

"""

# revision identifiers, used by Alembic.
revision = "0.56.4"
down_revision = "2d201872e23c"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""JSON database backup engine."""

import json
import os
import re
import shutil
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Generator,
    TextIO,
)

from sqlalchemy import Engine, MetaData, func, text
from sqlalchemy.schema import CreateIndex, CreateTable
from sqlmodel import (
    select,
)

from zenml.logger import get_logger
from zenml.utils.json_utils import pydantic_encoder
from zenml.zen_stores.migrations.backup.base import BaseDatabaseBackupEngine

logger = get_logger(__name__)

if TYPE_CHECKING:
    from zenml.zen_stores.sql_zen_store import SqlZenStoreConfiguration


class SQLAlchemyDatabaseBackupEngine(BaseDatabaseBackupEngine):
    """Base class for SQLAlchemy-based database backup engines."""

    def backup_database_to_storage(
        self,
        **store_db_kwargs: Any,
    ) -> None:
        """Backup the database to a storage location.

        Backup the database to an abstract storage location. The storage
        location is implemented by the `store_database_data` method that is
        called repeatedly to store the database information.

        Args:
            store_db_kwargs: Additional keyword arguments to pass to the
                `store_database_data` method.
        """
        metadata = MetaData()
        metadata.reflect(bind=self.engine)
        with self.engine.connect() as conn:
            for table in metadata.sorted_tables:
                # 1. extract the table creation statements

                create_table_construct = CreateTable(table)
                create_table_stmt = str(create_table_construct).strip()
                for column in create_table_construct.columns:
                    # enclosing all column names in backticks. This is because
                    # some column names are reserved keywords in MySQL. For
                    # example, keys and values. So, instead of tracking all
                    # keywords, we just enclose all column names in backticks.
                    # enclose the first word in the column definition in
                    # backticks
                    words = str(column).split()
                    words[0] = f"`{words[0]}`"
                    create_table_stmt = create_table_stmt.replace(
                        f"\n\t{str(column)}", " ".join(words)
                    )
                # if any double quotes are used for column names, replace them
                # with backticks
                create_table_stmt = create_table_stmt.replace('"', "") + ";"

                # enclose all table names in backticks. This is because some
                # table names are reserved keywords in MySQL (e.g key
                # and trigger).
                create_table_stmt = create_table_stmt.replace(
                    f"CREATE TABLE {table.name}",
                    f"CREATE TABLE `{table.name}`",
                )
                # do the same for references to other tables
                # (i.e. foreign key constraints) by replacing REFERENCES <word>
                # with REFERENCES `<word>`
                # use a regular expression for this
                create_table_stmt = re.sub(
                    r"REFERENCES\s+(\w+)",
                    r"REFERENCES `\1`",
                    create_table_stmt,
                )

                # In SQLAlchemy, the CreateTable statement may not always
                # include unique constraints explicitly if they are implemented
                # as unique indexes instead. To make sure we get all unique
                # constraints, including those implemented as indexes, we
                # extract the unique constraints from the table schema and add
                # them to the create table statement.

                # Extract the unique constraints from the table schema
                index_create_statements = []
                unique_constraints = []
                for index in table.indexes:
                    if index.unique:
                        unique_columns = [
                            f"`{column.name}`" for column in index.columns
                        ]
                        unique_constraints.append(
                            f"UNIQUE KEY `{index.name}` ({', '.join(unique_columns)})"
                        )
                    else:
                        if index.name in {
                            fk.name for fk in table.foreign_key_constraints
                        }:
                            # Foreign key indices are already handled by the
                            # table creation statement.
                            continue

                        index_create = str(CreateIndex(index)).strip()
                        index_create = index_create.replace(
                            f"CREATE INDEX {index.name}",
                            f"CREATE INDEX `{index.name}`",
                        )
                        index_create = index_create.replace(
                            f"ON {table.name}", f"ON `{table.name}`"
                        )

                        for column_name in index.columns.keys():
                            # We need this logic here to avoid the column names
                            # inside the index name
                            index_create = index_create.replace(
                                f"({column_name}", f"(`{column_name}`"
                            )
                            index_create = index_create.replace(
                                f"{column_name},", f"`{column_name}`,"
                            )
                            index_create = index_create.replace(
                                f"{column_name})", f"`{column_name}`)"
                            )

                        index_create = index_create.replace('"', "") + ";"
                        index_create_statements.append(index_create)

                # Add the unique constraints to the create table statement
                if unique_constraints:
                    # Remove the closing parenthesis, semicolon and any
                    # whitespaces at the end of the create table statement
                    create_table_stmt = re.sub(
                        r"\s*\)\s*;\s*$", "", create_table_stmt
                    )
                    create_table_stmt = (
                        create_table_stmt
                        + ", \n\t"
                        + ", \n\t".join(unique_constraints)
                        + "\n);"
                    )

                # Detect self-referential foreign keys from the table schema
                has_self_referential_foreign_keys = False
                for fk in table.foreign_keys:
                    # Check if the foreign key points to the same table
                    if fk.column.table == table:
                        has_self_referential_foreign_keys = True
                        break

                # Store the table schema
                self.store_database_data(
                    dict(
                        table=table.name,
                        create_stmt=create_table_stmt,
                        self_references=has_self_referential_foreign_keys,
                    ),
                    **store_db_kwargs,
                )

                for stmt in index_create_statements:
                    self.store_database_data(
                        dict(
                            table=table.name,
                            index_create_stmt=stmt,
                        ),
                        **store_db_kwargs,
                    )

                # 2. extract the table data in batches
                order_by = [col for col in table.primary_key]

                # Fetch the number of rows in the table
                row_count = conn.scalar(
                    select(func.count()).select_from(table)
                )

                # Fetch the data from the table in batches
                if row_count is not None:
                    batch_size = 100
                    for i in range(0, row_count, batch_size):
                        rows = conn.execute(
                            table.select()
                            .order_by(*order_by)
                            .limit(batch_size)
                            .offset(i)
                        ).fetchall()

                        self.store_database_data(
                            dict(
                                table=table.name,
                                data=[row._asdict() for row in rows],
                            ),
                            **store_db_kwargs,
                        )

    def restore_database_from_storage(
        self,
        **load_db_kwargs: Any,
    ) -> None:
        """Restore the database from a backup storage location.

        Restores the database from an abstract storage location. The storage
        location is implemented by the `load_database_data` method that is
        called repeatedly to load the database information from the external
        storage chunk by chunk.

        Args:
            load_db_kwargs: Additional keyword arguments to pass to the
                `load_database_data` method.
        """
        # Drop and re-create the primary database
        self.create_database(drop=True)

        metadata = MetaData()

        with self.engine.begin() as connection:
            # read the DB information one JSON object at a time
            self_references: dict[str, bool] = {}
            for table_dump in self.load_database_data(**load_db_kwargs):
                table_name = table_dump["table"]
                if "create_stmt" in table_dump:
                    # execute the table creation statement
                    connection.execute(text(table_dump["create_stmt"]))
                    # Reload the database metadata after creating the table
                    metadata.reflect(bind=self.engine)
                    self_references[table_name] = table_dump.get(
                        "self_references", False
                    )

                if "index_create_stmt" in table_dump:
                    # execute the index creation statement
                    connection.execute(text(table_dump["index_create_stmt"]))
                    # Reload the database metadata after creating the index
                    metadata.reflect(bind=self.engine)

                if "data" in table_dump:
                    # insert the data into the database
                    table = metadata.tables[table_name]
                    if self_references.get(table_name, False):
                        # If the table has self-referential foreign keys, we
                        # need to disable the foreign key checks before inserting
                        # the rows and re-enable them afterwards. This is because
                        # the rows need to be inserted in the correct order to
                        # satisfy the foreign key constraints and we don't sort
                        # the rows by creation time in the backup.
                        connection.execute(text("SET FOREIGN_KEY_CHECKS = 0"))

                    for row in table_dump["data"]:
                        # Convert column values to the correct type
                        for column in table.columns:
                            # Blob columns are stored as binary strings
                            if column.type.python_type is bytes and isinstance(
                                row[column.name], str
                            ):
                                # Convert the string to bytes
                                row[column.name] = bytes(
                                    row[column.name], "utf-8"
                                )

                    # Insert the rows into the table in batches
                    batch_size = 100
                    for i in range(0, len(table_dump["data"]), batch_size):
                        connection.execute(
                            table.insert().values(
                                table_dump["data"][i : i + batch_size]
                            )
                        )

                    if table_dump.get("self_references", False):
                        # Re-enable the foreign key checks after inserting the rows
                        connection.execute(text("SET FOREIGN_KEY_CHECKS = 1"))

    @abstractmethod
    def store_database_data(self, data: dict[str, Any], **kwargs: Any) -> None:
        """Store the database data.

        This method is called repeatedly to store the database information. It
        is called with a single argument, which is a dictionary containing
        either the table schema or table data. The dictionary contains the
        following keys:

            * `table`: The name of the table.
            * `create_stmt`: The table creation statement.
            * `data`: A list of rows in the table.

        The format of the dump is as depicted in the following example:

        ```json
        {
            "table": "table1",
            "create_stmt": "CREATE TABLE table1 (id INTEGER NOT NULL, "
                "name VARCHAR(255), PRIMARY KEY (id))"
        }
        {
            "table": "table1",
            "data": [
            {
                "id": 1,
                "name": "foo"
            },
            {
                "id": 1,
                "name": "bar"
            },
            ...
            ]
        }
        {
            "table": "table1",
            "data": [
            {
                "id": 101,
                "name": "fee"
            },
            {
                "id": 102,
                "name": "bee"
            },
            ...
            ]
        }
        ```

        Args:
            data: The database data to store.
            kwargs: Additional keyword arguments passed to the method.
        """

    @abstractmethod
    def load_database_data(
        self, **kwargs: Any
    ) -> Generator[dict[str, Any], None, None]:
        """Generator that loads the database data.

        This method is called repeatedly to load the database data. It must
        yield a dictionary containing either the table schema or table data,
        as documented in the `store_database_data` method.

        Yields:
            The loaded database data.

        Args:
            kwargs: Additional keyword arguments passed to the method.
        """


class InMemoryDatabaseBackupEngine(SQLAlchemyDatabaseBackupEngine):
    """In-memory database backup engine."""

    def __init__(
        self,
        config: "SqlZenStoreConfiguration",
        location: str | None = None,
    ) -> None:
        """Initialize the in-memory database backup engine.

        Args:
            config: The configuration of the store.
            location: The custom location to store the backup.
        """
        super().__init__(config, location or "memory")
        self.database_data: list[dict[str, Any]] = []

    def store_database_data(self, data: dict[str, Any], **kwargs: Any) -> None:
        """Store the database data.

        Args:
            data: The database data to store.
            kwargs: Additional keyword arguments passed to the method.
        """
        self.database_data.append(data)

    def load_database_data(
        self, **kwargs: Any
    ) -> Generator[dict[str, Any], None, None]:
        """Generator that loads the database data.

        Yields:
            The loaded database data.

        Args:
            kwargs: Additional keyword arguments passed to the method.
        """
        for data in self.database_data:
            yield data

    def backup_database(
        self,
        overwrite: bool = False,
    ) -> None:
        """Backup the database.

        Args:
            overwrite: Whether to overwrite an existing backup if it exists.
                If set to False, the existing backup will be reused.
        """
        if len(self.database_data) > 0:
            if not overwrite:
                logger.warning(
                    "An existing backup already exists. Reusing the existing backup."
                )
                return

            self.cleanup_database_backup()

        self.backup_database_to_storage()

        logger.debug("Database backed up to memory.")

    def restore_database(
        self,
        cleanup: bool = False,
    ) -> None:
        """Restore the database.

        Args:
            cleanup: Whether to cleanup the backup after restoring the database.

        Raises:
            RuntimeError: If no in-memory backup exists.
        """
        if len(self.database_data) == 0:
            raise RuntimeError(
                "No in-memory backup exists. Please backup the database first."
            )

        self.restore_database_from_storage()

        logger.debug("Database restored from memory.")

        if cleanup:
            self.cleanup_database_backup()

    def cleanup_database_backup(
        self,
    ) -> None:
        """Delete the database backup."""
        self.database_data = []

        logger.debug("In-memory database backup cleaned up.")


class FileDatabaseBackupEngine(SQLAlchemyDatabaseBackupEngine):
    """Database backup engine that stores the database data in a file."""

    def __init__(
        self,
        config: "SqlZenStoreConfiguration",
        location: str | None = None,
    ) -> None:
        """Initialize the file database backup engine.

        Args:
            config: The configuration of the store.
            location: The custom location to store the backup.
        """
        super().__init__(config, location)
        if self._backup_location is None:
            self._backup_location = self._get_db_backup_file_path()

    def _get_db_backup_file_path(self) -> str:
        """Get the path to the database backup.

        Returns:
            The path to the configured database backup.
        """
        from zenml.zen_stores.sql_zen_store import SQLDatabaseDriver

        assert self.url.database is not None

        if self.config.driver == SQLDatabaseDriver.SQLITE:
            return os.path.join(
                self.config.backup_directory,
                # Add the -backup suffix to the database filename
                self.url.database[:-3] + "-backup.db",
            )

        return os.path.join(
            self.config.backup_directory,
            f"{self.url.database}-backup.json",
        )

    def store_database_data(self, data: dict[str, Any], **kwargs: Any) -> None:
        """Store the database data.

        Args:
            data: The database data to store.
            **kwargs: Must include `dump_file` (TextIO) - the file handle
                to store the database data.
        """
        dump_file: TextIO = kwargs["dump_file"]
        # Write the data to the JSON file. Use an encoder that
        # can handle datetime, Decimal and other types.
        json.dump(
            data,
            dump_file,
            indent=4,
            default=pydantic_encoder,
        )
        dump_file.write("\n")

    def load_database_data(
        self, **kwargs: Any
    ) -> Generator[dict[str, Any], None, None]:
        """Generator that loads the database data.

        Args:
            **kwargs: Must include `dump_file` (TextIO) - the file handle
                to load the database data from.

        Yields:
            The loaded database data.
        """
        dump_file: TextIO = kwargs["dump_file"]
        buffer = ""
        while True:
            chunk = dump_file.readline()
            if not chunk:
                break
            buffer += chunk
            if chunk.rstrip() == "}":
                yield json.loads(buffer)
                buffer = ""

    def backup_database(
        self,
        overwrite: bool = False,
    ) -> None:
        """Backup the database.

        Args:
            overwrite: Whether to overwrite an existing backup if it exists.
                If set to False, the existing backup will be reused.
        """
        if os.path.isfile(self.backup_location):
            if not overwrite:
                logger.warning(
                    f"A previous backup file already exists at `{self.backup_location}`. "
                    "Reusing the existing backup."
                )
                return

            self.cleanup_database_backup()

        if self.url.drivername == "sqlite":
            # For a sqlite database, we can just make a copy of the database
            # file
            assert self.url.database is not None
            shutil.copyfile(
                self.url.database,
                self.backup_location,
            )
            return

        with open(self.backup_location, "w") as f:
            self.backup_database_to_storage(dump_file=f)

        logger.debug(f"Database backed up to file `{self.backup_location}`.")

    def restore_database(
        self,
        cleanup: bool = False,
    ) -> None:
        """Restore the database.

        Args:
            cleanup: Whether to cleanup the backup after restoring the database.

        Raises:
            RuntimeError: If the backup file does not exist or is not accessible.
        """
        if not os.path.isfile(self.backup_location):
            raise RuntimeError(
                f"Database backup file `{self.backup_location}` does not "
                "exist or is not accessible."
            )

        if self.url.drivername == "sqlite":
            # For a sqlite database, we just overwrite the database file
            # with the backup file
            assert self.url.database is not None
            shutil.copyfile(
                self.backup_location,
                self.url.database,
            )
            return

        with open(self.backup_location, "r") as f:
            self.restore_database_from_storage(dump_file=f)

        logger.debug(f"Database restored from file `{self.backup_location}`.")

        if cleanup:
            self.cleanup_database_backup()

    def cleanup_database_backup(
        self,
    ) -> None:
        """Delete the database backup."""
        if os.path.isfile(self.backup_location):
            try:
                os.remove(self.backup_location)
            except OSError:
                logger.warning(
                    f"Failed to cleanup database dump file `{self.backup_location}`."
                )
            else:
                logger.info(
                    f"Successfully cleaned up database dump file "
                    f"`{self.backup_location}`."
                )


class DBCloneDatabaseBackupEngine(BaseDatabaseBackupEngine):
    """Database backup engine that copies the database to a new database."""

    def __init__(
        self,
        config: "SqlZenStoreConfiguration",
        location: str | None = None,
    ) -> None:
        """Initialize the database backup engine.

        Args:
            config: The configuration of the store.
            location: The custom location to store the backup.

        Raises:
            ValueError: If the backup database name is not set in the store
                configuration.
        """
        super().__init__(config, location)
        if self._backup_location is None:
            if self.config.backup_database is None:
                raise ValueError(
                    "The backup database name must be set in the store "
                    "configuration to use the backup database strategy."
                )
            self._backup_location = self.config.backup_database

    @classmethod
    def _copy_database(cls, src_engine: Engine, dst_engine: Engine) -> None:
        """Copy the database from one engine to another.

        This method assumes that the destination database exists and is empty.

        Args:
            src_engine: The source SQLAlchemy engine.
            dst_engine: The destination SQLAlchemy engine.
        """
        src_metadata = MetaData()
        src_metadata.reflect(bind=src_engine)

        dst_metadata = MetaData()
        dst_metadata.reflect(bind=dst_engine)

        # @event.listens_for(src_metadata, "column_reflect")
        # def generalize_datatypes(inspector, tablename, column_dict):
        #     column_dict["type"] = column_dict["type"].as_generic(allow_nulltype=True)

        # Create all tables in the target database
        for table in src_metadata.sorted_tables:
            table.create(bind=dst_engine)

        # Refresh target metadata after creating the tables
        dst_metadata.clear()
        dst_metadata.reflect(bind=dst_engine)

        # Copy all data from the source database to the destination database
        with src_engine.begin() as src_conn:
            with dst_engine.begin() as dst_conn:
                for src_table in src_metadata.sorted_tables:
                    dst_table = dst_metadata.tables[src_table.name]
                    insert = dst_table.insert()

                    # If the table has a `created` column, we use it to sort
                    # the rows in the table starting with the oldest rows.
                    # This is to ensure that the rows are inserted in the
                    # correct order, since some tables have inner foreign key
                    # constraints.
                    if "created" in src_table.columns:
                        order_by = [src_table.columns["created"]]
                    else:
                        order_by = []
                    if "id" in src_table.columns:
                        # If the table has an `id` column, we also use it to
                        # sort the rows in the table, even if we already use
                        # "created" to sort the rows. We need a unique field to
                        # sort the rows, to break the tie between rows with the
                        # same "created" date, otherwise the same entry might
                        # end up multiple times in subsequent pages.
                        order_by.append(src_table.columns["id"])

                    row_count = src_conn.scalar(
                        select(func.count()).select_from(src_table)
                    )

                    # Copy rows in batches
                    if row_count is not None:
                        batch_size = 50
                        for i in range(0, row_count, batch_size):
                            rows = src_conn.execute(
                                src_table.select()
                                .order_by(*order_by)
                                .limit(batch_size)
                                .offset(i)
                            ).fetchall()

                            dst_conn.execute(
                                insert, [row._asdict() for row in rows]
                            )

    def backup_database(
        self,
        overwrite: bool = False,
    ) -> None:
        """Backup the database.

        Args:
            overwrite: Whether to overwrite an existing backup if it exists.
                If set to False, the existing backup will be reused.
        """
        if self.database_exists(database=self.backup_location):
            if not overwrite:
                logger.warning(
                    f"A previous backup database already exists at "
                    f"`{self.backup_location}`. Reusing the existing backup."
                )
                return

            self.cleanup_database_backup()

        self.create_database(
            database=self.backup_location,
            drop=True,
        )

        backup_engine = self.create_engine(database=self.backup_location)

        self._copy_database(self.engine, backup_engine)

        logger.debug(
            f"Database backed up to the `{self.backup_location}` backup database."
        )

    def restore_database(
        self,
        cleanup: bool = False,
    ) -> None:
        """Restore the database.

        Args:
            cleanup: Whether to cleanup the backup after restoring the database.

        Raises:
            RuntimeError: If the backup database does not exist.
        """
        if not self.database_exists(database=self.backup_location):
            raise RuntimeError(
                f"Backup database `{self.backup_location}` does not exist."
            )

        backup_engine = self.create_engine(database=self.backup_location)

        self.create_database(drop=True)

        self._copy_database(backup_engine, self.engine)

        logger.debug(
            f"Database restored from the `{self.backup_location}` backup database."
        )

        if cleanup:
            self.cleanup_database_backup()

    def cleanup_database_backup(
        self,
    ) -> None:
        """Delete the database backup."""
        if self.database_exists(database=self.backup_location):
            self.drop_database(
                database=self.backup_location,
            )
            logger.debug(
                f"Successfully cleaned up backup database "
                f"{self.backup_location}."
            )



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""CLI functionality to interact with authorized devices."""

from typing import Any

import click

from zenml.cli import utils as cli_utils
from zenml.cli.cli import TagGroup, cli
from zenml.cli.utils import OutputFormat, list_options
from zenml.client import Client
from zenml.console import console
from zenml.enums import CliCategories
from zenml.logger import get_logger
from zenml.models import OAuthDeviceFilter

logger = get_logger(__name__)


@cli.group(cls=TagGroup, tag=CliCategories.MANAGEMENT_TOOLS)
def authorized_device() -> None:
    """Interact with authorized devices."""


@authorized_device.command("describe")
@click.argument("id_or_prefix", type=str, required=True)
def describe_authorized_device(id_or_prefix: str) -> None:
    """Fetch an authorized device.

    Args:
        id_or_prefix: The ID of the authorized device to fetch.
    """
    try:
        device = Client().get_authorized_device(
            id_or_prefix=id_or_prefix,
        )
    except KeyError as e:
        cli_utils.exception(e)

    cli_utils.print_pydantic_model(
        title=f"Authorized device `{device.id}`",
        model=device,
        exclude_columns={"user"},
    )


@authorized_device.command(
    "list", help="List all authorized devices for the current user."
)
@list_options(
    OAuthDeviceFilter,
    default_columns=[
        "id",
        "status",
        "expires",
        "hostname",
        "os",
    ],
)
def list_authorized_devices(
    columns: str, output_format: OutputFormat, **kwargs: Any
) -> None:
    """List all authorized devices.

    Args:
        columns: Columns to display in output.
        output_format: Format for output (table/json/yaml/csv/tsv).
        **kwargs: Keyword arguments to filter authorized devices.
    """
    with console.status("Listing authorized devices...\n"):
        devices = Client().list_authorized_devices(**kwargs)

    cli_utils.print_page(
        devices,
        columns,
        output_format,
        empty_message="No authorized devices found for this filter.",
    )


@authorized_device.command("lock")
@click.argument("id", type=str, required=True)
def lock_authorized_device(id: str) -> None:
    """Lock an authorized device.

    Args:
        id: The ID of the authorized device to lock.
    """
    try:
        Client().update_authorized_device(
            id_or_prefix=id,
            locked=True,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Locked authorized device `{id}`.")


@authorized_device.command("unlock")
@click.argument("id", type=str, required=True)
def unlock_authorized_device(id: str) -> None:
    """Unlock an authorized device.

    Args:
        id: The ID of the authorized device to unlock.
    """
    try:
        Client().update_authorized_device(
            id_or_prefix=id,
            locked=False,
        )
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Locked authorized device `{id}`.")


@authorized_device.command("delete")
@click.argument("id", type=str, required=True)
@click.option(
    "--yes",
    "-y",
    is_flag=True,
    help="Don't ask for confirmation.",
)
def delete_authorized_device(id: str, yes: bool = False) -> None:
    """Delete an authorized device.

    Args:
        id: The ID of the authorized device to delete.
        yes: If set, don't ask for confirmation.
    """
    if not yes:
        confirmation = cli_utils.confirmation(
            f"Are you sure you want to delete authorized device `{id}`?"
        )
        if not confirmation:
            cli_utils.declare("Authorized device deletion canceled.")
            return

    try:
        Client().delete_authorized_device(id_or_prefix=id)
    except KeyError as e:
        cli_utils.exception(e)
    else:
        cli_utils.declare(f"Deleted authorized device `{id}`.")



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML HyperAI orchestrator."""

from typing import TYPE_CHECKING, Dict, Optional, Type

from pydantic import Field

from zenml.config.base_settings import BaseSettings
from zenml.integrations.hyperai import HYPERAI_RESOURCE_TYPE
from zenml.logger import get_logger
from zenml.models import ServiceConnectorRequirements
from zenml.orchestrators import (
    BaseOrchestratorConfig,
    BaseOrchestratorFlavor,
)

if TYPE_CHECKING:
    from zenml.integrations.hyperai.orchestrators import HyperAIOrchestrator

logger = get_logger(__name__)


class HyperAIOrchestratorSettings(BaseSettings):
    """HyperAI orchestrator settings."""

    mounts_from_to: Dict[str, str] = Field(
        default_factory=dict,
        description="A dictionary mapping from paths on the HyperAI instance "
        "to paths within the Docker container. This allows users to mount "
        "directories from the HyperAI instance into the Docker container.",
    )


class HyperAIOrchestratorConfig(
    BaseOrchestratorConfig, HyperAIOrchestratorSettings
):
    """Configuration for the HyperAI orchestrator."""

    container_registry_autologin: bool = Field(
        False,
        description="If True, the orchestrator will attempt to automatically "
        "log in to the container registry specified in the stack configuration "
        "on the HyperAI instance.",
    )
    automatic_cleanup_pipeline_files: bool = Field(
        True,
        description="If True, the orchestrator will automatically clean up old "
        "pipeline files that are on the HyperAI instance. Pipeline files will "
        "be cleaned up if they are 7 days old or older.",
    )
    gpu_enabled_in_container: bool = Field(
        True,
        description="If True, the orchestrator will enable GPU support in the "
        "Docker container that runs on the HyperAI instance.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return True

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return True


class HyperAIOrchestratorFlavor(BaseOrchestratorFlavor):
    """Flavor for the HyperAI orchestrator."""

    @property
    def name(self) -> str:
        """Name of the orchestrator flavor.

        Returns:
            Name of the orchestrator flavor.
        """
        return "hyperai"

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type=HYPERAI_RESOURCE_TYPE
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/connectors/hyperai/hyperai.png"

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return HyperAIOrchestratorConfig

    @property
    def implementation_class(self) -> Type["HyperAIOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        from zenml.integrations.hyperai.orchestrators import (
            HyperAIOrchestrator,
        )

        return HyperAIOrchestrator



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implements the log stores for ZenML."""

# Base classes
from zenml.log_stores.base_log_store import (
    BaseLogStore,
    BaseLogStoreConfig,
    BaseLogStoreFlavor,
)

# OpenTelemetry log store
from zenml.log_stores.otel.otel_flavor import (
    OtelLogStoreConfig,
    OtelLogStoreFlavor,
)
from zenml.log_stores.otel.otel_log_store import OtelLogStore

# Artifact log store
from zenml.log_stores.artifact.artifact_log_store import (
    ArtifactLogStore,
)

# Datadog log store
from zenml.log_stores.datadog.datadog_flavor import (
    DatadogLogStoreConfig,
    DatadogLogStoreFlavor,
)
from zenml.log_stores.datadog.datadog_log_store import (
    DatadogLogStore,
)

__all__ = [
    "ArtifactLogStore",
    "BaseLogStore",
    "BaseLogStoreConfig",
    "BaseLogStoreFlavor",
    "DatadogLogStore",
    "DatadogLogStoreConfig",
    "DatadogLogStoreFlavor",
    "OtelLogStore",
    "OtelLogStoreConfig",
    "OtelLogStoreFlavor",
]



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Skypilot orchestrator Lambda flavor."""

from typing import TYPE_CHECKING, Any, Optional, Type

from zenml.integrations.skypilot.flavors.skypilot_orchestrator_base_vm_config import (
    SkypilotBaseOrchestratorConfig,
    SkypilotBaseOrchestratorSettings,
)
from zenml.integrations.skypilot_lambda import (
    SKYPILOT_LAMBDA_ORCHESTRATOR_FLAVOR,
)
from zenml.logger import get_logger
from zenml.orchestrators import BaseOrchestratorConfig, BaseOrchestratorFlavor
from zenml.utils.secret_utils import SecretField

if TYPE_CHECKING:
    from zenml.integrations.skypilot_lambda.orchestrators import (
        SkypilotLambdaOrchestrator,
    )


logger = get_logger(__name__)


class SkypilotLambdaOrchestratorSettings(SkypilotBaseOrchestratorSettings):
    """Skypilot orchestrator settings."""

    _UNSUPPORTED_FEATURES = {
        "use_spot": "Spot instances not supported for Lambda orchestrator.",
        "job_recovery": "Job recovery not supported for Lambda orchestrator.",
        "image_id": "Custom image IDs not supported for Lambda orchestrator.",
        # Add other unsupported features as needed
    }

    def __setattr__(self, name: str, value: Any) -> None:
        """Set attribute.

        Args:
            name: Name of the attribute.
            value: Value of the attribute.

        Raises:
            AttributeError: If the attribute is not supported.
        """
        if name in self._UNSUPPORTED_FEATURES:
            raise AttributeError(f"{name} is not supported on Lambda.")
        super().__setattr__(name, value)


class SkypilotLambdaOrchestratorConfig(
    SkypilotBaseOrchestratorConfig, SkypilotLambdaOrchestratorSettings
):
    """Skypilot orchestrator config."""

    api_key: Optional[str] = SecretField(default=None)


class SkypilotLambdaOrchestratorFlavor(BaseOrchestratorFlavor):
    """Flavor for the Skypilot Lambda orchestrator."""

    @property
    def name(self) -> str:
        """Name of the orchestrator flavor.

        Returns:
            Name of the orchestrator flavor.
        """
        return SKYPILOT_LAMBDA_ORCHESTRATOR_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the orchestrator flavor.

        Returns:
            Display name of the orchestrator flavor.
        """
        return "Skypilot Lambda"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/lambda.png"

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return SkypilotLambdaOrchestratorConfig

    @property
    def implementation_class(self) -> Type["SkypilotLambdaOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            Implementation class for this flavor.
        """
        from zenml.integrations.skypilot_lambda.orchestrators import (
            SkypilotLambdaOrchestrator,
        )

        return SkypilotLambdaOrchestrator



"""Release [0.93.2].

Revision ID: 0.93.2
Revises: d1e8c2c4a9ef
Create Date: 2026-01-29 10:14:59.243786

"""

# revision identifiers, used by Alembic.
revision = "0.93.2"
down_revision = "d1e8c2c4a9ef"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Step context class."""

import contextvars
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
)

from typing_extensions import Self

from zenml.exceptions import StepContextError
from zenml.logger import get_logger
from zenml.utils import context_utils
from zenml.utils.callback_registry import CallbackRegistry
from zenml.utils.singleton import SingletonMetaClass

if TYPE_CHECKING:
    from zenml.artifacts.artifact_config import ArtifactConfig
    from zenml.materializers.base_materializer import BaseMaterializer
    from zenml.metadata.metadata_types import MetadataType
    from zenml.model.model import Model
    from zenml.models import (
        PipelineResponse,
        PipelineRunResponse,
        StepRunResponse,
    )
    from zenml.models.v2.core.step_run import StepRunInputResponse


logger = get_logger(__name__)


def get_step_context() -> "StepContext":
    """Get the context of the currently running step.

    Returns:
        The context of the currently running step.

    Raises:
        RuntimeError: If no step is currently running.
    """
    if ctx := StepContext.get():
        return ctx

    raise RuntimeError(
        "The step context is only available inside a step function."
    )


def get_or_create_run_context() -> "RunContext":
    """Get or create the context of the currently running pipeline.

    Returns:
        The context of the currently running pipeline.
    """
    return RunContext()


class RunContext(metaclass=SingletonMetaClass):
    """Provides context shared between all steps in a pipeline run."""

    def __init__(self) -> None:
        """Create the run context."""
        self.initialized = False
        self._state: Optional[Any] = None

    @property
    def state(self) -> Optional[Any]:
        """Returns the pipeline state.

        Returns:
            The pipeline state or None.

        Raises:
            RuntimeError: If the run context is not initialized.
        """
        if not self.initialized:
            raise RuntimeError(
                "Run context not initialized. The run state is only available "
                "in the context of a running pipeline."
            )
        return self._state

    def initialize(self, state: Optional[Any]) -> None:
        """Initialize the run context.

        Args:
            state: Optional state for the pipeline run

        Raises:
            RuntimeError: If the run context is already initialized.
        """
        if self.initialized:
            raise RuntimeError("Run context already initialized.")
        self._state = state
        self.initialized = True


class StepContext(context_utils.BaseContext):
    """Provides additional context inside a step function.

    Usage example:

    ```python
    from zenml.steps import get_step_context

    @step
    def my_trainer_step() -> Any:
        context = get_step_context()

        # get info about the current pipeline run
        current_pipeline_run = context.pipeline_run

        # get info about the current step run
        current_step_run = context.step_run

        # get info about the future output artifacts of this step
        output_artifact_uri = context.get_output_artifact_uri()

        ...
    ```
    """

    __context_var__ = contextvars.ContextVar("step_context")

    def __init__(
        self,
        pipeline_run: "PipelineRunResponse",
        step_run: "StepRunResponse",
        output_materializers: Mapping[str, Sequence[Type["BaseMaterializer"]]],
        output_artifact_uris: Mapping[str, str],
        output_artifact_configs: Mapping[str, Optional["ArtifactConfig"]],
    ) -> None:
        """Initialize the context of the currently running step.

        Args:
            pipeline_run: The model of the current pipeline run.
            step_run: The model of the current step run.
            output_materializers: The output materializers of the step that
                this context is used in.
            output_artifact_uris: The output artifacts of the step that this
                context is used in.
            output_artifact_configs: The outputs' ArtifactConfigs of the step that this
                context is used in.

        Raises:
            StepContextError: If the keys of the output materializers and
                output artifacts do not match.
        """
        from zenml.client import Client

        super().__init__()

        try:
            pipeline_run = Client().get_pipeline_run(pipeline_run.id)
        except KeyError:
            pass
        self.pipeline_run = pipeline_run
        try:
            step_run = Client().get_run_step(step_run.id)
        except KeyError:
            pass
        self.step_run = step_run
        self.model_version = (
            step_run.model_version or pipeline_run.model_version
        )

        self.step_name = self.step_run.name

        # set outputs
        if output_materializers.keys() != output_artifact_uris.keys():
            raise StepContextError(
                f"Mismatched keys in output materializers and output artifact "
                f"URIs for step `{self.step_name}`. Output materializer "
                f"keys: {set(output_materializers)}, output artifact URI "
                f"keys: {set(output_artifact_uris)}"
            )
        self._outputs = {
            key: StepContextOutput(
                materializer_classes=output_materializers[key],
                artifact_uri=output_artifact_uris[key],
                artifact_config=output_artifact_configs[key],
            )
            for key in output_materializers.keys()
        }
        self._cleanup_registry = CallbackRegistry()

    @property
    def pipeline(self) -> "PipelineResponse":
        """Returns the current pipeline.

        Returns:
            The current pipeline or None.

        Raises:
            StepContextError: If the pipeline run does not have a pipeline.
        """
        if self.pipeline_run.pipeline:
            return self.pipeline_run.pipeline
        raise StepContextError(
            f"Unable to get pipeline in step `{self.step_name}` of pipeline "
            f"run '{self.pipeline_run.id}': This pipeline run does not have "
            f"a pipeline associated with it."
        )

    @property
    def pipeline_state(self) -> Optional[Any]:
        """Returns the pipeline state.

        Returns:
            The pipeline state or None.
        """
        return get_or_create_run_context().state

    @property
    def model(self) -> "Model":
        """Returns configured Model.

        Order of resolution to search for Model is:
            1. Model from the step context
            2. Model from the pipeline context

        Returns:
            The `Model` object associated with the current step.

        Raises:
            StepContextError: If no `Model` object was specified for the step
                or pipeline.
        """
        if not self.model_version:
            raise StepContextError(
                f"Unable to get Model in step `{self.step_name}` of pipeline "
                f"run '{self.pipeline_run.id}': No model has been specified "
                "the step or pipeline."
            )

        return self.model_version.to_model_class()

    @property
    def inputs(self) -> Dict[str, List["StepRunInputResponse"]]:
        """Returns the input artifacts of the current step.

        Returns:
            The input artifacts of the current step.
        """
        return self.step_run.regular_inputs

    def _get_output(
        self, output_name: Optional[str] = None
    ) -> "StepContextOutput":
        """Returns the materializer and artifact URI for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                materializer and URI.

        Returns:
            Tuple containing the materializer and artifact URI for the
                given output.

        Raises:
            StepContextError: If the step has no outputs, no output for
                the given `output_name` or if no `output_name` was given but
                the step has multiple outputs.
        """
        output_count = len(self._outputs)
        if output_count == 0:
            raise StepContextError(
                f"Unable to get step output for step `{self.step_name}`: "
                f"This step does not have any outputs."
            )

        if not output_name and output_count > 1:
            raise StepContextError(
                f"Unable to get step output for step `{self.step_name}`: "
                f"This step has multiple outputs ({set(self._outputs)}), "
                f"please specify which output to return."
            )

        if output_name:
            if output_name not in self._outputs:
                raise StepContextError(
                    f"Unable to get step output '{output_name}' for "
                    f"step `{self.step_name}`. This step does not have an "
                    f"output with the given name, please specify one of the "
                    f"available outputs: {set(self._outputs)}."
                )
            return self._outputs[output_name]
        else:
            return next(iter(self._outputs.values()))

    def get_output_materializer(
        self,
        output_name: Optional[str] = None,
        custom_materializer_class: Optional[Type["BaseMaterializer"]] = None,
        data_type: Optional[Type[Any]] = None,
    ) -> "BaseMaterializer":
        """Returns a materializer for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                materializer. If no name is given and the step only has a
                single output, the materializer of this output will be
                returned. If the step has multiple outputs, an exception
                will be raised.
            custom_materializer_class: If given, this `BaseMaterializer`
                subclass will be initialized with the output artifact instead
                of the materializer that was registered for this step output.
            data_type: If the output annotation is of type `Union` and the step
                therefore has multiple materializers configured, you can provide
                a data type for the output which will be used to select the
                correct materializer. If not provided, the first materializer
                will be used.

        Returns:
            A materializer initialized with the output artifact for
            the given output.
        """
        from zenml.utils import materializer_utils

        output = self._get_output(output_name)
        materializer_classes = output.materializer_classes
        artifact_uri = output.artifact_uri

        if custom_materializer_class:
            materializer_class = custom_materializer_class
        elif len(materializer_classes) == 1 or not data_type:
            materializer_class = materializer_classes[0]
        else:
            materializer_class = materializer_utils.select_materializer(
                data_type=data_type, materializer_classes=materializer_classes
            )

        return materializer_class(artifact_uri)

    def get_output_artifact_uri(
        self, output_name: Optional[str] = None
    ) -> str:
        """Returns the artifact URI for a given step output.

        Args:
            output_name: Optional name of the output for which to get the URI.
                If no name is given and the step only has a single output,
                the URI of this output will be returned. If the step has
                multiple outputs, an exception will be raised.

        Returns:
            Artifact URI for the given output.
        """
        return self._get_output(output_name).artifact_uri

    def get_output_metadata(
        self, output_name: Optional[str] = None
    ) -> Dict[str, "MetadataType"]:
        """Returns the metadata for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                metadata. If no name is given and the step only has a single
                output, the metadata of this output will be returned. If the
                step has multiple outputs, an exception will be raised.

        Returns:
            Metadata for the given output.
        """
        output = self._get_output(output_name)
        custom_metadata = output.run_metadata or {}
        if output.artifact_config:
            custom_metadata.update(
                **(output.artifact_config.run_metadata or {})
            )
        return custom_metadata

    def get_output_tags(self, output_name: Optional[str] = None) -> List[str]:
        """Returns the tags for a given step output.

        Args:
            output_name: Optional name of the output for which to get the
                metadata. If no name is given and the step only has a single
                output, the metadata of this output will be returned. If the
                step has multiple outputs, an exception will be raised.

        Returns:
            Tags for the given output.
        """
        output = self._get_output(output_name)
        custom_tags = set(output.tags or [])
        if output.artifact_config:
            return list(
                set(output.artifact_config.tags or []).union(custom_tags)
            )
        return list(custom_tags)

    def add_output_metadata(
        self,
        metadata: Dict[str, "MetadataType"],
        output_name: Optional[str] = None,
    ) -> None:
        """Adds metadata for a given step output.

        Args:
            metadata: The metadata to add.
            output_name: Optional name of the output for which to add the
                metadata. If no name is given and the step only has a single
                output, the metadata of this output will be added. If the
                step has multiple outputs, an exception will be raised.
        """
        output = self._get_output(output_name)
        if not output.run_metadata:
            output.run_metadata = {}
        output.run_metadata.update(**metadata)

    def add_output_tags(
        self,
        tags: List[str],
        output_name: Optional[str] = None,
    ) -> None:
        """Adds tags for a given step output.

        Args:
            tags: The tags to add.
            output_name: Optional name of the output for which to add the
                tags. If no name is given and the step only has a single
                output, the tags of this output will be added. If the
                step has multiple outputs, an exception will be raised.
        """
        output = self._get_output(output_name)
        if not output.tags:
            output.tags = []
        output.tags += tags

    def remove_output_tags(
        self,
        tags: List[str],
        output_name: Optional[str] = None,
    ) -> None:
        """Removes tags for a given step output.

        Args:
            tags: The tags to remove.
            output_name: Optional name of the output for which to remove the
                tags. If no name is given and the step only has a single
                output, the tags of this output will be removed. If the
                step has multiple outputs, an exception will be raised.
        """
        output = self._get_output(output_name)
        if not output.tags:
            return
        output.tags = [tag for tag in output.tags if tag not in tags]

    def __enter__(self) -> Self:
        """Enter the step context.

        Raises:
            RuntimeError: If the step context has already been entered.

        Returns:
            The step context object.
        """
        if self._token is not None:
            raise RuntimeError(
                "Running a step from within another step is not allowed."
            )
        return super().__enter__()

    def __exit__(self, *_: Any) -> None:
        """Exit the step context.

        Args:
            *_: Unused keyword arguments.
        """
        self._cleanup_registry.execute_callbacks(raise_on_exception=False)
        super().__exit__(*_)


class StepContextOutput:
    """Represents a step output in the step context."""

    materializer_classes: Sequence[Type["BaseMaterializer"]]
    artifact_uri: str
    run_metadata: Optional[Dict[str, "MetadataType"]] = None
    artifact_config: Optional["ArtifactConfig"]
    tags: Optional[List[str]] = None

    def __init__(
        self,
        materializer_classes: Sequence[Type["BaseMaterializer"]],
        artifact_uri: str,
        artifact_config: Optional["ArtifactConfig"],
    ):
        """Initialize the step output.

        Args:
            materializer_classes: The materializer classes for the output.
            artifact_uri: The artifact URI for the output.
            artifact_config: The ArtifactConfig object of the output.
        """
        self.materializer_classes = materializer_classes
        self.artifact_uri = artifact_uri
        self.artifact_config = artifact_config



"""add pipeline deployment [0d69e308846a].

Revision ID: 0d69e308846a
Revises: 8ad841ad9bfe
Create Date: 2025-08-26 10:30:52.737833

"""

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = "0d69e308846a"
down_revision = "8ad841ad9bfe"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "deployment",
        sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("created", sa.DateTime(), nullable=False),
        sa.Column("updated", sa.DateTime(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("project_id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
        sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
        sa.Column(
            "status", sqlmodel.sql.sqltypes.AutoString(), nullable=False
        ),
        sa.Column("url", sa.TEXT(), nullable=True),
        sa.Column("auth_key", sa.TEXT(), nullable=True),
        sa.Column(
            "deployment_metadata",
            sa.String(length=16777215).with_variant(mysql.MEDIUMTEXT, "mysql"),
            nullable=False,
        ),
        sa.Column(
            "snapshot_id",
            sqlmodel.sql.sqltypes.GUID(),
            nullable=True,
        ),
        sa.Column("deployer_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
        sa.ForeignKeyConstraint(
            ["snapshot_id"],
            ["pipeline_snapshot.id"],
            name="fk_deployment_snapshot_id_pipeline_snapshot",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["deployer_id"],
            ["stack_component.id"],
            name="fk_deployment_deployer_id_stack_component",
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_deployment_project_id_project",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name="fk_deployment_user_id_user",
            ondelete="SET NULL",
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint(
            "name",
            "project_id",
            name="unique_deployment_name_in_project",
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("deployment")
    # ### end Alembic commands ###
